from pytest_cases import fixture_ref
from pytest_cases import parametrize, fixture_ref

from labelbox.data.metrics.confusion_matrix.confusion_matrix import confusion_matrix_metric


@parametrize("tool_examples", [
    fixture_ref('polygon_pairs'),
    fixture_ref('rectangle_pairs'),
    fixture_ref('mask_pairs'),
    fixture_ref('line_pairs'),
    fixture_ref('point_pairs'),
    fixture_ref('ner_pairs')
])
def test_overlapping_objects(tool_examples):
    for example in tool_examples:

        for include_subclasses, expected_attr_name in [[
                True, 'expected'
        ], [False, 'expected_without_subclasses']]:
            score = confusion_matrix_metric(
                example.ground_truths,
                example.predictions,
                include_subclasses=include_subclasses)

            if len(getattr(example, expected_attr_name)) == 0:
                assert len(score) == 0
            else:
                expected = [0, 0, 0, 0]
                for expected_values in getattr(example,
                                               expected_attr_name).values():
                    for idx in range(4):
                        expected[idx] += expected_values[idx]
                assert score[0].value == tuple(
                    expected), f"{example.predictions},{example.ground_truths}"


@parametrize("tool_examples",
             [fixture_ref('checklist_pairs'),
              fixture_ref('radio_pairs')])
def test_overlapping_classifications(tool_examples):
    for example in tool_examples:
        score = confusion_matrix_metric(example.ground_truths,
                                        example.predictions)
        if len(example.expected) == 0:
            assert len(score) == 0
        else:
            expected = [0, 0, 0, 0]
            for expected_values in example.expected.values():
                for idx in range(4):
                    expected[idx] += expected_values[idx]
            assert score[0].value == tuple(
                expected), f"{example.predictions},{example.ground_truths}"


def test_partial_overlap(pair_iou_thresholds):
    for example in pair_iou_thresholds:
        for iou in example.expected.keys():
            score = confusion_matrix_metric(example.predictions,
                                            example.ground_truths,
                                            iou=iou)
            assert score[0].value == tuple(
                example.expected[iou]
            ), f"{example.predictions},{example.ground_truths}"



from email import message
import pytest
from pydantic import ValidationError

from labelbox.data.annotation_types.data.generic_data_row_data import GenericDataRowData
from labelbox.data.annotation_types.data.video import VideoData
from labelbox.data.annotation_types.label import Label


def test_generic_data_type():
    data = {
        'global_key':
            'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr',
    }
    label = Label(data=data)
    data = label.data
    assert isinstance(data, GenericDataRowData)
    assert data.global_key == 'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr'


def test_generic_data_type_validations():
    data = {
        'row_data':
            'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr',
    }
    with pytest.raises(ValueError, match="Exactly one of"):
        Label(data=data)

    data = {
        'uid':
            "abcd",
        'global_key':
            'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr',
    }
    with pytest.raises(ValueError, match="Only one of"):
        Label(data=data)


def test_video_data_type():
    data = {
        'global_key':
            'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr',
    }
    with pytest.warns(UserWarning, match="Use a dict"):
        label = Label(data=VideoData(**data))
    data = label.data
    assert isinstance(data, VideoData)
    assert data.global_key == 'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr'


def test_generic_data_row():
    data = {
        'global_key':
            'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr',
    }
    label = Label(data=GenericDataRowData(**data))
    data = label.data
    assert isinstance(data, GenericDataRowData)
    assert data.global_key == 'https://lb-test-data.s3.us-west-1.amazonaws.com/image-samples/sample-image-1.jpg-BEidMVWRmyXjVCnr'



from enum import Enum
from typing import Optional

from labelbox.schema.media_type import MediaType


class OntologyKind(Enum):
    """
    OntologyKind is an enum that represents the different types of ontologies
    At the moment it is only limited to ModelEvaluation
    """
    ModelEvaluation = "MODEL_EVALUATION"
    Missing = None

    @classmethod
    def is_supported(cls, value):
        return isinstance(value, cls)

    @classmethod
    def get_ontology_kind_validation_error(cls, ontology_kind):
        return TypeError(f"{ontology_kind}: is not a valid ontology kind. Use"
                         f" any of {OntologyKind.__members__.items()}"
                         " from OntologyKind.")


class EditorTaskType(Enum):
    ModelChatEvaluation = "MODEL_CHAT_EVALUATION"
    ResponseCreation = "RESPONSE_CREATION"
    OfflineModelChatEvaluation = "OFFLINE_MODEL_CHAT_EVALUATION"
    Missing = None

    @classmethod
    def is_supported(cls, value):
        return isinstance(value, cls)

    @classmethod
    def _missing_(cls, value) -> 'EditorTaskType':
        """Handle missing null new task types
            Handle upper case names for compatibility with
            the GraphQL"""

        if value is None:
            return cls.Missing

        for name, member in cls.__members__.items():
            if value == name.upper():
                return member

        return cls.Missing


class EditorTaskTypeMapper:

    @staticmethod
    def to_editor_task_type(ontology_kind: OntologyKind,
                            media_type: MediaType) -> EditorTaskType:
        if ontology_kind and OntologyKind.is_supported(
                ontology_kind) and media_type and MediaType.is_supported(
                    media_type):
            editor_task_type = EditorTaskTypeMapper.map_to_editor_task_type(
                ontology_kind, media_type)
        else:
            editor_task_type = EditorTaskType.Missing

        return editor_task_type

    @staticmethod
    def map_to_editor_task_type(onotology_kind: OntologyKind,
                                media_type: MediaType) -> EditorTaskType:
        if onotology_kind == OntologyKind.ModelEvaluation and media_type == MediaType.Conversational:
            return EditorTaskType.ModelChatEvaluation
        else:
            return EditorTaskType.Missing


class UploadType(Enum):
    Auto = 'AUTO',
    Manual = 'MANUAL',
    Missing = None

    @classmethod
    def is_supported(cls, value):
        return isinstance(value, cls)

    @classmethod
    def _missing_(cls, value: object) -> 'UploadType':
        if value is None:
            return cls.Missing

        for name, member in cls.__members__.items():
            if value == name.upper():
                return member

        return cls.Missing



import pytest
from collections import defaultdict
from unittest.mock import MagicMock
from labelbox import Client
from labelbox.exceptions import ResourceConflict, ResourceCreationError, ResourceNotFoundError, MalformedQueryException, UnprocessableEntityError
from labelbox.schema.project import Project
from labelbox.schema.user import User
from labelbox.schema.user_group import UserGroup, UserGroupColor
from labelbox.schema.queue_mode import QueueMode
from labelbox.schema.ontology_kind import EditorTaskType
from labelbox.schema.media_type import MediaType

@pytest.fixture
def group_user():
    user_values = defaultdict(lambda: None)
    user_values["id"] = "user_id"
    user_values["email"] = "test@example.com"
    return User(MagicMock(Client), user_values)


@pytest.fixture
def group_project():
    project_values = defaultdict(lambda: None)
    project_values["id"] = "project_id"
    project_values["name"] = "Test Project"
    project_values["queueMode"] = QueueMode.Batch.value
    project_values["editorTaskType"] = EditorTaskType.Missing.value
    project_values["mediaType"] = MediaType.Image.value
    return Project(MagicMock(Client), project_values)


class TestUserGroupColor:

    def test_user_group_color_values(self):
        assert UserGroupColor.BLUE.value == "9EC5FF"
        assert UserGroupColor.PURPLE.value == "CEB8FF"
        assert UserGroupColor.ORANGE.value == "FFB35F"
        assert UserGroupColor.CYAN.value == "4ED2F9"
        assert UserGroupColor.PINK.value == "FFAEA9"
        assert UserGroupColor.LIGHT_PINK.value == "FFA9D5"
        assert UserGroupColor.GREEN.value == "3FDC9A"
        assert UserGroupColor.YELLOW.value == "E7BF00"
        assert UserGroupColor.GRAY.value == "B8C4D3"


class TestUserGroup:

    def setup_method(self):
        self.client = MagicMock(Client)
        self.client.enable_experimental = True
        self.group = UserGroup(client=self.client)
  
    def test_constructor_experimental_needed(self):
        client = MagicMock(Client)
        client.enable_experimental = False
        with pytest.raises(RuntimeError):
            group = UserGroup(client)

    def test_constructor(self):
        group = UserGroup(self.client)

        assert group.id == ""
        assert group.name == ""
        assert group.color is UserGroupColor.BLUE
        assert len(group.projects) == 0
        assert len(group.users) == 0

    def test_update_with_exception_name(self):
        group = self.group
        group.id = ""

        with pytest.raises(ValueError):
            group.get()

    def test_get(self):
        projects = [
            {
                "id": "project_id_1",
                "name": "project_1"
            },
            {
                "id": "project_id_2",
                "name": "project_2"
            }
        ]
        group_members = [
            {
                "id": "user_id_1",
                "email": "email_1"
            },
            {
                "id": "user_id_2",
                "email": "email_2"
            }
        ]
        self.client.execute.return_value = {
            "userGroup": {
                "id": "group_id",
                "name": "Test Group",
                "color": "4ED2F9",
                "projects": {
                    "nodes": projects
                },
                "members": {
                    "nodes": group_members
                }
            }
        }
        group = UserGroup(self.client)
        assert group.id == ""
        assert group.name == ""
        assert group.color is UserGroupColor.BLUE
        assert len(group.projects) == 0
        assert len(group.users) == 0

        group.id = "group_id"
        group.get()

        assert group.id == "group_id"
        assert group.name == "Test Group"
        assert group.color is UserGroupColor.CYAN
        assert len(group.projects) == 2
        assert len(group.users) == 2

    def test_get_value_error(self):
        self.client.execute.return_value = None
        group = UserGroup(self.client)
        group.name = "Test Group"

        with pytest.raises(ValueError):
            group.get()

    def test_update(self, group_user, group_project):
        group = self.group
        group.id = "group_id"
        group.name = "Test Group"
        group.color = UserGroupColor.BLUE
        group.users = { group_user }
        group.projects = { group_project }

        updated_group = group.update()

        execute = self.client.execute.call_args[0]

        assert "UpdateUserGroupPyApi" in execute[0]
        assert execute[1]["id"] == "group_id"
        assert execute[1]["name"] == "Test Group"
        assert execute[1]["color"] == UserGroupColor.BLUE.value
        assert len(execute[1]["userIds"]) == 1
        assert list(execute[1]["userIds"])[0] == group_user.uid
        assert len(execute[1]["projectIds"]) == 1
        assert list(execute[1]["projectIds"])[0] == group_project.uid

        assert updated_group.id == "group_id"
        assert updated_group.name == "Test Group"
        assert updated_group.color == UserGroupColor.BLUE
        assert len(updated_group.users) == 1
        assert list(updated_group.users)[0].uid == group_user.uid
        assert len(updated_group.projects) == 1
        assert list(updated_group.projects)[0].uid == group_project.uid

    def test_update_resource_error_input_bad(self):
        self.client.execute.side_effect = MalformedQueryException("Error")
        group = UserGroup(self.client)
        group.name = "Test Group"
        group.id = "group_id"

        with pytest.raises(UnprocessableEntityError):
            group.update()

    def test_update_resource_error_unknown_id(self):
        self.client.execute.return_value = None
        group = UserGroup(self.client)
        group.name = "Test Group"
        group.id = "group_id"

        with pytest.raises(ResourceNotFoundError) as e:
            group.update()

    def test_update_with_exception_name(self):
        group = self.group
        group.name = ""

        with pytest.raises(UnprocessableEntityError):
            group.update()

    def test_update_with_exception_name(self):
        group = self.group
        group.id = ""

        with pytest.raises(ValueError):
            group.update()

    def test_create_with_exception_id(self):
        group = self.group
        group.id = "group_id"

        with pytest.raises(ResourceCreationError):
            group.create()

    def test_create_with_exception_name(self):
        group = self.group
        group.name = ""

        with pytest.raises(ValueError):
            group.create()

    def test_create(self, group_user, group_project):
        group = self.group
        group.name = "New Group"
        group.color = UserGroupColor.PINK
        group.users = { group_user }
        group.projects = { group_project }

        self.client.execute.return_value = {
            "createUserGroup": {
                "group": {
                    "id": "group_id"
                }
            }
        }
        created_group = group.create()
        execute = self.client.execute.call_args[0]

        assert "CreateUserGroupPyApi" in execute[0]
        assert execute[1]["name"] == "New Group"
        assert execute[1]["color"] == UserGroupColor.PINK.value
        assert len(execute[1]["userIds"]) == 1
        assert list(execute[1]["userIds"])[0] == "user_id"
        assert len(execute[1]["projectIds"]) == 1
        assert list(execute[1]["projectIds"])[0] == "project_id"
        assert created_group.id is not None
        assert created_group.id == "group_id"
        assert created_group.name == "New Group"
        assert created_group.color == UserGroupColor.PINK
        assert len(created_group.users) == 1
        assert list(created_group.users)[0].uid == "user_id"
        assert len(created_group.projects) == 1
        assert list(created_group.projects)[0].uid == "project_id"
    
    def test_create_resource_creation_error(self):
        self.client.execute.side_effect = ResourceConflict("Error")
        group = UserGroup(self.client)
        group.name = "Test Group"

        with pytest.raises(ResourceCreationError):
            group.create()

    def test_delete(self):
        group = self.group
        group.id = "group_id"

        self.client.execute.return_value = {
            "deleteUserGroup": {
                "success": True
            }
        }
        deleted = group.delete()
        execute = self.client.execute.call_args[0]

        assert "DeleteUserGroupPyApi" in execute[0]
        assert execute[1]["id"] == "group_id"
        assert deleted is True

    def test_delete_resource_not_found_error(self):
        self.client.execute.return_value = None
        group = UserGroup(self.client)
        group.id = "group_id"

        with pytest.raises(ResourceNotFoundError):
            group.delete()

    def test_delete_no_id(self):
        group = UserGroup(self.client)
        group.id = None

        with pytest.raises(ValueError):
            group.delete()

    def test_user_groups_empty(self):
        self.client.execute.return_value = {"userGroups": None}

        user_groups = list(UserGroup(self.client).get_user_groups())

        assert len(user_groups) == 0

    def test_user_groups(self):
        self.client.execute.return_value = {
            "userGroups": {
                "nextCursor":
                    None,
                "nodes": [{
                    "id": "group_id_1",
                    "name": "Group 1",
                    "color": "9EC5FF",
                    "projects": {
                        "nodes": [{
                            "id": "project_id_1",
                            "name": "Project 1"
                        }, {
                            "id": "project_id_2",
                            "name": "Project 2"
                        }]
                    },
                    "members": {
                        "nodes": [{
                            "id": "user_id_1",
                            "email": "user1@example.com"
                        }, {
                            "id": "user_id_2",
                            "email": "user2@example.com"
                        }]
                    }
                }, {
                    "id": "group_id_2",
                    "name": "Group 2",
                    "color": "9EC5FF",
                    "projects": {
                        "nodes": [{
                            "id": "project_id_3",
                            "name": "Project 3"
                        }, {
                            "id": "project_id_4",
                            "name": "Project 4"
                        }]
                    },
                    "members": {
                        "nodes": [{
                            "id": "user_id_3",
                            "email": "user3@example.com"
                        }, {
                            "id": "user_id_4",
                            "email": "user4@example.com"
                        }]
                    }
                }, {
                    "id": "group_id_3",
                    "name": "Group 3",
                    "color": "9EC5FF",
                    "projects": {
                        "nodes": [{
                            "id": "project_id_5",
                            "name": "Project 5"
                        }, {
                            "id": "project_id_6",
                            "name": "Project 6"
                        }]
                    },
                    "members": {
                        "nodes": [{
                            "id": "user_id_5",
                            "email": "user5@example.com"
                        }, {
                            "id": "user_id_6",
                            "email": "user6@example.com"
                        }]
                    }
                }]
            }
        }

        user_groups = list(UserGroup(self.client).get_user_groups())
        execute = self.client.execute.call_args[0]

        assert "GetUserGroupsPyApi" in execute[0]
        assert len(user_groups) == 3

        # Check the attributes of the first user group
        assert user_groups[0].id == "group_id_1"
        assert user_groups[0].name == "Group 1"
        assert user_groups[0].color == UserGroupColor.BLUE
        assert len(user_groups[0].projects) == 2
        assert len(user_groups[0].users) == 2

        # Check the attributes of the second user group
        assert user_groups[1].id == "group_id_2"
        assert user_groups[1].name == "Group 2"
        assert user_groups[1].color == UserGroupColor.BLUE
        assert len(user_groups[1].projects) == 2
        assert len(user_groups[1].users) == 2

        # Check the attributes of the third user group
        assert user_groups[2].id == "group_id_3"
        assert user_groups[2].name == "Group 3"
        assert user_groups[2].color == UserGroupColor.BLUE
        assert len(user_groups[2].projects) == 2
        assert len(user_groups[2].users) == 2


if __name__ == "__main__":
    import subprocess
    subprocess.call(["pytest", "-v", __file__])



import uuid
import pytest
from unittest.mock import MagicMock, patch

from labelbox.schema.annotation_import import LabelImport, logger


def test_should_warn_user_about_unsupported_confidence():
    """this test should check running state only to validate running, not completed"""
    id = str(uuid.uuid4())

    labels = [
        {
            "uuid": "b862c586-8614-483c-b5e6-82810f70cac0",
            "schemaId": "ckrazcueb16og0z6609jj7y3y",
            "dataRow": {
                "id": "ckrazctum0z8a0ybc0b0o0g0v"
            },
            "confidence": 0.851,
            "bbox": {
                "top": 1352,
                "left": 2275,
                "height": 350,
                "width": 139
            }
        },
    ]
    with patch.object(LabelImport, '_create_label_import_from_bytes'):
        with patch.object(logger, 'warning') as warning_mock:
            LabelImport.create_from_objects(client=MagicMock(),
                                            project_id=id,
                                            name=id,
                                            labels=labels)
            warning_mock.assert_called_once()
            "Confidence scores are not supported in Label Import" in warning_mock.call_args_list[
                0].args[0]


def test_invalid_labels_format():
    """this test should confirm that labels are required to be in a form of list"""
    id = str(uuid.uuid4())

    label = {
        "uuid": "b862c586-8614-483c-b5e6-82810f70cac0",
        "schemaId": "ckrazcueb16og0z6609jj7y3y",
        "dataRow": {
            "id": "ckrazctum0z8a0ybc0b0o0g0v"
        },
        "bbox": {
            "top": 1352,
            "left": 2275,
            "height": 350,
            "width": 139
        }
    }
    with patch.object(LabelImport, '_create_label_import_from_bytes'):
        with pytest.raises(TypeError):
            LabelImport.create_from_objects(client=MagicMock(),
                                            project_id=id,
                                            name=id,
                                            labels=label)



import json
import pytest

from labelbox.data.serialization.ndjson.classification import NDRadio

from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.data.serialization.ndjson.objects import NDLine


def round_dict(data):
    if isinstance(data, dict):
        for key in data:
            if isinstance(data[key], float):
                data[key] = int(data[key])
            elif isinstance(data[key], dict):
                data[key] = round_dict(data[key])
            elif isinstance(data[key], (list, tuple)):
                data[key] = [round_dict(r) for r in data[key]]

    return data


@pytest.mark.parametrize('filename', [
    'tests/data/assets/ndjson/classification_import_global_key.json',
    'tests/data/assets/ndjson/metric_import_global_key.json',
    'tests/data/assets/ndjson/polyline_import_global_key.json',
    'tests/data/assets/ndjson/text_entity_import_global_key.json',
    'tests/data/assets/ndjson/conversation_entity_import_global_key.json',
])
def test_many_types(filename: str):
    with open(filename, 'r') as f:
        data = json.load(f)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data
    f.close()


def test_image():
    with open('tests/data/assets/ndjson/image_import_global_key.json',
              'r') as f:
        data = json.load(f)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    for r in res:
        r.pop('classifications', None)
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]
    f.close()


def test_pdf():
    with open('tests/data/assets/ndjson/pdf_import_global_key.json', 'r') as f:
        data = json.load(f)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert [round_dict(x) for x in res] == [round_dict(x) for x in data]
    f.close()


def test_video():
    with open('tests/data/assets/ndjson/video_import_global_key.json',
              'r') as f:
        data = json.load(f)

    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == [data[2], data[0], data[1], data[3], data[4], data[5]]
    f.close()



from labelbox.data.annotation_types.data.tiled_image import TiledImageData
from labelbox.utils import camel_case
from typing import List, Optional, Union, Dict, Any

from labelbox import pydantic_compat

from ...annotation_types.annotation import (ClassificationAnnotation,
                                            ObjectAnnotation)
from ...annotation_types.video import VideoClassificationAnnotation, VideoObjectAnnotation
from ...annotation_types.data import ImageData, TextData, VideoData
from ...annotation_types.label import Label
from .classification import LBV1Classifications
from .objects import LBV1Objects, LBV1TextEntity


class LBV1LabelAnnotations(LBV1Classifications, LBV1Objects):

    def to_common(
            self) -> List[Union[ObjectAnnotation, ClassificationAnnotation]]:
        classifications = LBV1Classifications.to_common(self)
        objects = LBV1Objects.to_common(self)
        return [*objects, *classifications]

    @classmethod
    def from_common(
        cls, annotations: List[Union[ClassificationAnnotation,
                                     ObjectAnnotation]]
    ) -> "LBV1LabelAnnotations":

        objects = LBV1Objects.from_common(
            [x for x in annotations if isinstance(x, ObjectAnnotation)])
        classifications = LBV1Classifications.from_common(
            [x for x in annotations if isinstance(x, ClassificationAnnotation)])
        return cls(**objects.dict(), **classifications.dict())


class LBV1LabelAnnotationsVideo(LBV1LabelAnnotations):
    frame_number: int = pydantic_compat.Field(..., alias='frameNumber')

    def to_common(
        self
    ) -> List[Union[VideoClassificationAnnotation, VideoObjectAnnotation]]:
        classifications = [
            VideoClassificationAnnotation(
                value=classification.to_common(),
                frame=self.frame_number,
                name=classification.title,
                feature_schema_id=classification.schema_id)
            for classification in self.classifications
        ]

        objects = [
            VideoObjectAnnotation(value=obj.to_common(),
                                  keyframe=obj.keyframe,
                                  classifications=[
                                      ClassificationAnnotation(
                                          value=cls.to_common(),
                                          feature_schema_id=cls.schema_id,
                                          name=cls.title,
                                          extra={
                                              'feature_id': cls.feature_id,
                                              'title': cls.title,
                                              'value': cls.value,
                                          }) for cls in obj.classifications
                                  ],
                                  name=obj.title,
                                  frame=self.frame_number,
                                  alternative_name=obj.value,
                                  feature_schema_id=obj.schema_id,
                                  extra={
                                      'value': obj.value,
                                      'instanceURI': obj.instanceURI,
                                      'color': obj.color,
                                      'feature_id': obj.feature_id,
                                  }) for obj in self.objects
        ]
        return [*classifications, *objects]

    @classmethod
    def from_common(
        cls, annotations: List[Union[VideoObjectAnnotation,
                                     VideoClassificationAnnotation]]
    ) -> "LBV1LabelAnnotationsVideo":
        by_frames = {}
        for annotation in annotations:
            if annotation.frame in by_frames:
                by_frames[annotation.frame].append(annotation)
            else:
                by_frames[annotation.frame] = [annotation]

        result = []
        for frame in by_frames:
            converted = LBV1LabelAnnotations.from_common(
                annotations=by_frames[frame])
            result.append(
                LBV1LabelAnnotationsVideo(
                    frame_number=frame,
                    objects=converted.objects,
                    classifications=converted.classifications))

        return result

    class Config:
        allow_population_by_field_name = True


class Review(pydantic_compat.BaseModel):
    score: int
    id: str
    created_at: str
    created_by: str
    label_id: Optional[str] = None

    class Config:
        alias_generator = camel_case


Extra = lambda name: pydantic_compat.Field(None, alias=name, extra_field=True)


class LBV1Label(pydantic_compat.BaseModel):
    label: Union[LBV1LabelAnnotations,
                 List[LBV1LabelAnnotationsVideo]] = pydantic_compat.Field(
                     ..., alias='Label')
    data_row_id: str = pydantic_compat.Field(..., alias="DataRow ID")
    row_data: str = pydantic_compat.Field(None, alias="Labeled Data")
    id: Optional[str] = pydantic_compat.Field(None, alias='ID')
    external_id: Optional[str] = pydantic_compat.Field(None,
                                                       alias="External ID")
    data_row_media_attributes: Optional[Dict[str, Any]] = pydantic_compat.Field(
        {}, alias="Media Attributes")
    data_row_metadata: Optional[List[Dict[str, Any]]] = pydantic_compat.Field(
        [], alias="DataRow Metadata")

    created_by: Optional[str] = Extra('Created By')
    project_name: Optional[str] = Extra('Project Name')
    created_at: Optional[str] = Extra('Created At')
    updated_at: Optional[str] = Extra('Updated At')
    seconds_to_label: Optional[float] = Extra('Seconds to Label')
    agreement: Optional[float] = Extra('Agreement')
    benchmark_agreement: Optional[float] = Extra('Benchmark Agreement')
    benchmark_id: Optional[str] = Extra('Benchmark ID')
    dataset_name: Optional[str] = Extra('Dataset Name')
    reviews: Optional[List[Review]] = Extra('Reviews')
    label_url: Optional[str] = Extra('View Label')
    has_open_issues: Optional[float] = Extra('Has Open Issues')
    skipped: Optional[bool] = Extra('Skipped')
    media_type: Optional[str] = Extra('media_type')
    data_split: Optional[str] = Extra('Data Split')
    global_key: Optional[str] = Extra('Global Key')

    def to_common(self) -> Label:
        if isinstance(self.label, list):
            annotations = []
            for lbl in self.label:
                annotations.extend(lbl.to_common())
        else:
            annotations = self.label.to_common()

        return Label(data=self._data_row_to_common(),
                     uid=self.id,
                     annotations=annotations,
                     extra={
                         field.alias: getattr(self, field_name)
                         for field_name, field in self.__fields__.items()
                         if field.field_info.extra.get('extra_field')
                     })

    @classmethod
    def from_common(cls, label: Label):
        if isinstance(label.annotations[0],
                      (VideoObjectAnnotation, VideoClassificationAnnotation)):
            label_ = LBV1LabelAnnotationsVideo.from_common(label.annotations)
        else:
            label_ = LBV1LabelAnnotations.from_common(label.annotations)
        return LBV1Label(label=label_,
                         id=label.uid,
                         data_row_id=label.data.uid,
                         row_data=label.data.url,
                         external_id=label.data.external_id,
                         data_row_media_attributes=label.data.media_attributes,
                         data_row_metadata=label.data.metadata,
                         **label.extra)

    def _data_row_to_common(
            self) -> Union[ImageData, TextData, VideoData, TiledImageData]:
        # Use data row information to construct the appropriate annotation type
        data_row_info = {
            'url' if self._is_url() else 'text': self.row_data,
            'external_id': self.external_id,
            'uid': self.data_row_id,
            'media_attributes': self.data_row_media_attributes,
            'metadata': self.data_row_metadata
        }

        self.media_type = self.media_type or self._infer_media_type()
        media_mapping = {
            'text': TextData,
            'image': ImageData,
            'video': VideoData
        }
        if self.media_type not in media_mapping:
            raise ValueError(
                f"Annotation types are only supported for {list(media_mapping)} media types."
                f" Found {self.media_type}.")
        return media_mapping[self.media_type](**data_row_info)

    def _infer_media_type(self) -> str:
        # Determines the data row type based on the label content
        if isinstance(self.label, list):
            return 'video'
        if self._has_text_annotations():
            return 'text'
        elif self._has_object_annotations():
            return 'image'
        else:
            if self._row_contains((".jpg", ".png", ".jpeg")) and self._is_url():
                return 'image'
            elif (self._row_contains((".txt", ".text", ".html")) and
                  self._is_url()) or not self._is_url():
                return 'text'
            else:
                #  This condition will occur when a data row url does not contain a file extension
                #  and the label does not contain object annotations that indicate the media type.
                #  As a temporary workaround you can explicitly set the media_type
                #  in each label json payload before converting.
                #  We will eventually provide the media type in the export.
                raise TypeError(
                    f"Can't infer data type from row data. row_data: {self.row_data[:200]}"
                )

    def _has_object_annotations(self) -> bool:
        return len(self.label.objects) > 0

    def _has_text_annotations(self) -> bool:
        return len([
            annotation for annotation in self.label.objects
            if isinstance(annotation, LBV1TextEntity)
        ]) > 0

    def _row_contains(self, substrs) -> bool:
        lower_row_data = self.row_data.lower()
        return any([substr in lower_row_data for substr in substrs])

    def _is_url(self) -> bool:
        return self.row_data.startswith(
            ("http://", "https://", "gs://",
             "s3://")) or "tileLayerUrl" in self.row_data

    class Config:
        allow_population_by_field_name = True



import json

from unittest.mock import MagicMock, Mock
import pytest

from labelbox.schema.internal.descriptor_file_creator import DescriptorFileCreator


def test_chunk_down_by_bytes_row_too_large():
    client = MagicMock()

    descriptor_file_creator = DescriptorFileCreator(client)

    chunk = [{"row_data": "a"}]
    max_chunk_size_bytes = 1

    res = descriptor_file_creator._chunk_down_by_bytes(chunk,
                                                       max_chunk_size_bytes)
    assert [x for x in res] == [json.dumps([{"row_data": "a"}])]


def test_chunk_down_by_bytes_more_chunks():
    client = MagicMock()

    descriptor_file_creator = DescriptorFileCreator(client)

    chunk = [{"row_data": "a"}, {"row_data": "b"}]
    max_chunk_size_bytes = len(json.dumps(chunk).encode("utf-8")) - 1

    res = descriptor_file_creator._chunk_down_by_bytes(chunk,
                                                       max_chunk_size_bytes)
    assert [x for x in res] == [
        json.dumps([{
            "row_data": "a"
        }]), json.dumps([{
            "row_data": "b"
        }])
    ]


def test_chunk_down_by_bytes_one_chunk():
    client = MagicMock()

    descriptor_file_creator = DescriptorFileCreator(client)

    chunk = [{"row_data": "a"}, {"row_data": "b"}]
    max_chunk_size_bytes = len(json.dumps(chunk).encode("utf-8"))

    res = descriptor_file_creator._chunk_down_by_bytes(chunk,
                                                       max_chunk_size_bytes)
    assert [x for x in res
           ] == [json.dumps([{
               "row_data": "a"
           }, {
               "row_data": "b"
           }])]



import pytest
from unittest.mock import MagicMock, patch

from labelbox.schema.project import Project
from labelbox.schema.ontology_kind import EditorTaskType


@pytest.fixture
def project_entity():
    return Project(
        MagicMock(), {
            "id": "test",
            "name": "test",
            "createdAt": "2021-06-01T00:00:00.000Z",
            "updatedAt": "2021-06-01T00:00:00.000Z",
            "autoAuditNumberOfLabels": 1,
            "autoAuditPercentage": 100,
            "dataRowCount": 1,
            "description": "test",
            "editorTaskType": "MODEL_CHAT_EVALUATION",
            "lastActivityTime": "2021-06-01T00:00:00.000Z",
            "allowedMediaType": "IMAGE",
            "queueMode": "BATCH",
            "setupComplete": "2021-06-01T00:00:00.000Z",
            "modelSetupComplete": None,
            "uploadType": "Auto",
        })


@pytest.mark.parametrize(
    'api_editor_task_type, expected_editor_task_type',
    [(None, EditorTaskType.Missing),
     ('MODEL_CHAT_EVALUATION', EditorTaskType.ModelChatEvaluation),
     ('RESPONSE_CREATION', EditorTaskType.ResponseCreation),
     ('OFFLINE_MODEL_CHAT_EVALUATION',
      EditorTaskType.OfflineModelChatEvaluation),
     ('NEW_TYPE', EditorTaskType.Missing)])
def test_project_editor_task_type(api_editor_task_type,
                                  expected_editor_task_type, project_entity):
    client = MagicMock()
    project = Project(
        client, {
            "id": "test",
            "name": "test",
            "createdAt": "2021-06-01T00:00:00.000Z",
            "updatedAt": "2021-06-01T00:00:00.000Z",
            "autoAuditNumberOfLabels": 1,
            "autoAuditPercentage": 100,
            "dataRowCount": 1,
            "description": "test",
            "editorTaskType": api_editor_task_type,
            "lastActivityTime": "2021-06-01T00:00:00.000Z",
            "allowedMediaType": "IMAGE",
            "queueMode": "BATCH",
            "setupComplete": "2021-06-01T00:00:00.000Z",
            "modelSetupComplete": None,
            "uploadType": "Auto",
        })

    assert project.editor_task_type == expected_editor_task_type


def test_setup_editor_using_connect_ontology(project_entity):
    project = project_entity
    ontology = MagicMock()
    project.connect_ontology = MagicMock()
    with patch("warnings.warn") as warn:
        project.setup_editor(ontology)
        warn.assert_called_once()
        project.connect_ontology.assert_called_once_with(ontology)



from labelbox.typing_imports import Literal
from labelbox.utils import _NoCoercionMixin
from .base_data import BaseData


class LlmPromptResponseCreationData(BaseData, _NoCoercionMixin):
    class_name: Literal[
        "LlmPromptResponseCreationData"] = "LlmPromptResponseCreationData"


import json
from pathlib import Path

from labelbox.data.serialization.coco import COCOConverter

COCO_ASSETS_DIR = "tests/data/assets/coco"


def run_instances(tmpdir):
    instance_json = json.load(open(Path(COCO_ASSETS_DIR, 'instances.json')))
    res = COCOConverter.deserialize_instances(instance_json,
                                              Path(COCO_ASSETS_DIR, 'images'))
    back = COCOConverter.serialize_instances(
        res,
        Path(tmpdir),
    )


def test_rle_objects(tmpdir):
    rle_json = json.load(open(Path(COCO_ASSETS_DIR, 'rle.json')))
    res = COCOConverter.deserialize_instances(rle_json,
                                              Path(COCO_ASSETS_DIR, 'images'))
    back = COCOConverter.serialize_instances(res, tmpdir)


def test_panoptic(tmpdir):
    panoptic_json = json.load(open(Path(COCO_ASSETS_DIR, 'panoptic.json')))
    image_dir, mask_dir = [
        Path(COCO_ASSETS_DIR, dir_name) for dir_name in ['images', 'masks']
    ]
    res = COCOConverter.deserialize_panoptic(panoptic_json, image_dir, mask_dir)
    back = COCOConverter.serialize_panoptic(res,
                                            Path(f'/{tmpdir}/images_panoptic'),
                                            Path(f'/{tmpdir}/masks_panoptic'))



from io import BytesIO
from typing import Any, Dict, List, Tuple, Union, Optional
import base64

from labelbox.data.annotation_types.ner.conversation_entity import ConversationEntity
from labelbox.data.annotation_types.video import VideoObjectAnnotation, DICOMObjectAnnotation
from labelbox.data.mixins import ConfidenceMixin, CustomMetricsMixin, CustomMetric, CustomMetricsNotSupportedMixin
import numpy as np

from labelbox import pydantic_compat
from PIL import Image
from labelbox.data.annotation_types import feature

from labelbox.data.annotation_types.data.video import VideoData

from ...annotation_types.data import ImageData, TextData, MaskData
from ...annotation_types.ner import DocumentEntity, DocumentTextSelection, TextEntity
from ...annotation_types.types import Cuid
from ...annotation_types.geometry import DocumentRectangle, Rectangle, Polygon, Line, Point, Mask
from ...annotation_types.annotation import ClassificationAnnotation, ObjectAnnotation
from ...annotation_types.video import VideoMaskAnnotation, DICOMMaskAnnotation, MaskFrame, MaskInstance
from .classification import NDClassification, NDSubclassification, NDSubclassificationType
from .base import DataRow, NDAnnotation, NDJsonBase


class NDBaseObject(NDAnnotation):
    classifications: List[NDSubclassificationType] = []


class VideoSupported(pydantic_compat.BaseModel):
    # support for video for objects are per-frame basis
    frame: int


class DicomSupported(pydantic_compat.BaseModel):
    group_key: str


class _Point(pydantic_compat.BaseModel):
    x: float
    y: float


class Bbox(pydantic_compat.BaseModel):
    top: float
    left: float
    height: float
    width: float


class NDPoint(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    point: _Point

    def to_common(self) -> Point:
        return Point(x=self.point.x, y=self.point.y)

    @classmethod
    def from_common(
            cls,
            uuid: str,
            point: Point,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None) -> "NDPoint":
        return cls(point={
            'x': point.x,
            'y': point.y
        },
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDFramePoint(VideoSupported):
    point: _Point
    classifications: List[NDSubclassificationType] = []

    def to_common(self, name: str, feature_schema_id: Cuid,
                  segment_index: int) -> VideoObjectAnnotation:
        return VideoObjectAnnotation(frame=self.frame,
                                     segment_index=segment_index,
                                     keyframe=True,
                                     name=name,
                                     feature_schema_id=feature_schema_id,
                                     value=Point(x=self.point.x,
                                                 y=self.point.y),
                                     classifications=[
                                         NDSubclassification.to_common(annot)
                                         for annot in self.classifications
                                     ])

    @classmethod
    def from_common(
        cls,
        frame: int,
        point: Point,
        classifications: List[NDSubclassificationType],
    ):
        return cls(frame=frame,
                   point=_Point(x=point.x, y=point.y),
                   classifications=classifications)


class NDLine(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    line: List[_Point]

    def to_common(self) -> Line:
        return Line(points=[Point(x=pt.x, y=pt.y) for pt in self.line])

    @classmethod
    def from_common(
            cls,
            uuid: str,
            line: Line,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None) -> "NDLine":
        return cls(line=[{
            'x': pt.x,
            'y': pt.y
        } for pt in line.points],
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDFrameLine(VideoSupported):
    line: List[_Point]
    classifications: List[NDSubclassificationType] = []

    def to_common(self, name: str, feature_schema_id: Cuid,
                  segment_index: int) -> VideoObjectAnnotation:
        return VideoObjectAnnotation(
            frame=self.frame,
            segment_index=segment_index,
            keyframe=True,
            name=name,
            feature_schema_id=feature_schema_id,
            value=Line(points=[Point(x=pt.x, y=pt.y) for pt in self.line]),
            classifications=[
                NDSubclassification.to_common(annot)
                for annot in self.classifications
            ])

    @classmethod
    def from_common(
        cls,
        frame: int,
        line: Line,
        classifications: List[NDSubclassificationType],
    ):
        return cls(frame=frame,
                   line=[{
                       'x': pt.x,
                       'y': pt.y
                   } for pt in line.points],
                   classifications=classifications)


class NDDicomLine(NDFrameLine):

    def to_common(self, name: str, feature_schema_id: Cuid, segment_index: int,
                  group_key: str) -> DICOMObjectAnnotation:
        return DICOMObjectAnnotation(
            frame=self.frame,
            segment_index=segment_index,
            keyframe=True,
            name=name,
            feature_schema_id=feature_schema_id,
            value=Line(points=[Point(x=pt.x, y=pt.y) for pt in self.line]),
            group_key=group_key)


class NDPolygon(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    polygon: List[_Point]

    def to_common(self) -> Polygon:
        return Polygon(points=[Point(x=pt.x, y=pt.y) for pt in self.polygon])

    @classmethod
    def from_common(
            cls,
            uuid: str,
            polygon: Polygon,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None) -> "NDPolygon":
        return cls(polygon=[{
            'x': pt.x,
            'y': pt.y
        } for pt in polygon.points],
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDRectangle(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    bbox: Bbox

    def to_common(self) -> Rectangle:
        return Rectangle(start=Point(x=self.bbox.left, y=self.bbox.top),
                         end=Point(x=self.bbox.left + self.bbox.width,
                                   y=self.bbox.top + self.bbox.height))

    @classmethod
    def from_common(
            cls,
            uuid: str,
            rectangle: Rectangle,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None
    ) -> "NDRectangle":
        return cls(bbox=Bbox(top=min(rectangle.start.y, rectangle.end.y),
                             left=min(rectangle.start.x, rectangle.end.x),
                             height=abs(rectangle.end.y - rectangle.start.y),
                             width=abs(rectangle.end.x - rectangle.start.x)),
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   page=extra.get('page'),
                   unit=extra.get('unit'),
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDDocumentRectangle(NDRectangle):
    page: int
    unit: str

    def to_common(self) -> DocumentRectangle:
        return DocumentRectangle(start=Point(x=self.bbox.left, y=self.bbox.top),
                                 end=Point(x=self.bbox.left + self.bbox.width,
                                           y=self.bbox.top + self.bbox.height),
                                 page=self.page,
                                 unit=self.unit)

    @classmethod
    def from_common(
            cls,
            uuid: str,
            rectangle: DocumentRectangle,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None
    ) -> "NDRectangle":
        return cls(bbox=Bbox(top=min(rectangle.start.y, rectangle.end.y),
                             left=min(rectangle.start.x, rectangle.end.x),
                             height=abs(rectangle.end.y - rectangle.start.y),
                             width=abs(rectangle.end.x - rectangle.start.x)),
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   page=rectangle.page,
                   unit=rectangle.unit.value,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDFrameRectangle(VideoSupported):
    bbox: Bbox
    classifications: List[NDSubclassificationType] = []

    def to_common(self, name: str, feature_schema_id: Cuid,
                  segment_index: int) -> VideoObjectAnnotation:
        return VideoObjectAnnotation(
            frame=self.frame,
            segment_index=segment_index,
            keyframe=True,
            name=name,
            feature_schema_id=feature_schema_id,
            value=Rectangle(start=Point(x=self.bbox.left, y=self.bbox.top),
                            end=Point(x=self.bbox.left + self.bbox.width,
                                      y=self.bbox.top + self.bbox.height)),
            classifications=[
                NDSubclassification.to_common(annot)
                for annot in self.classifications
            ])

    @classmethod
    def from_common(
        cls,
        frame: int,
        rectangle: Rectangle,
        classifications: List[NDSubclassificationType],
    ):
        return cls(frame=frame,
                   bbox=Bbox(top=min(rectangle.start.y, rectangle.end.y),
                             left=min(rectangle.start.x, rectangle.end.x),
                             height=abs(rectangle.end.y - rectangle.start.y),
                             width=abs(rectangle.end.x - rectangle.start.x)),
                   classifications=classifications)


class NDSegment(pydantic_compat.BaseModel):
    keyframes: List[Union[NDFrameRectangle, NDFramePoint, NDFrameLine]]

    @staticmethod
    def lookup_segment_object_type(segment: List) -> "NDFrameObjectType":
        """Used for determining which object type the annotation contains
        returns the object type"""
        result = {
            Rectangle: NDFrameRectangle,
            Point: NDFramePoint,
            Line: NDFrameLine,
        }.get(type(segment[0].value))
        return result

    @staticmethod
    def segment_with_uuid(keyframe: Union[NDFrameRectangle, NDFramePoint,
                                          NDFrameLine], uuid: str):
        keyframe._uuid = uuid
        keyframe.extra = {'uuid': uuid}
        return keyframe

    def to_common(self, name: str, feature_schema_id: Cuid, uuid: str,
                  segment_index: int):
        return [
            self.segment_with_uuid(
                keyframe.to_common(name=name,
                                   feature_schema_id=feature_schema_id,
                                   segment_index=segment_index), uuid)
            for keyframe in self.keyframes
        ]

    @classmethod
    def from_common(cls, segment):
        nd_frame_object_type = cls.lookup_segment_object_type(segment)

        return cls(keyframes=[
            nd_frame_object_type.from_common(
                object_annotation.frame, object_annotation.value, [
                    NDSubclassification.from_common(annot)
                    for annot in object_annotation.classifications
                ])
            for object_annotation in segment
        ])


class NDDicomSegment(NDSegment):
    keyframes: List[NDDicomLine]

    @staticmethod
    def lookup_segment_object_type(segment: List) -> "NDDicomObjectType":
        """Used for determining which object type the annotation contains
        returns the object type"""
        segment_class = type(segment[0].value)
        if segment_class == Line:
            return NDDicomLine
        else:
            raise ValueError('DICOM segments only support Line objects')

    def to_common(self, name: str, feature_schema_id: Cuid, uuid: str,
                  segment_index: int, group_key: str):
        return [
            self.segment_with_uuid(
                keyframe.to_common(name=name,
                                   feature_schema_id=feature_schema_id,
                                   segment_index=segment_index,
                                   group_key=group_key), uuid)
            for keyframe in self.keyframes
        ]


class NDSegments(NDBaseObject):
    segments: List[NDSegment]

    def to_common(self, name: str, feature_schema_id: Cuid):
        result = []
        for idx, segment in enumerate(self.segments):
            result.extend(
                segment.to_common(name=name,
                                  feature_schema_id=feature_schema_id,
                                  segment_index=idx,
                                  uuid=self.uuid))
        return result

    @classmethod
    def from_common(cls, segments: List[VideoObjectAnnotation], data: VideoData,
                    name: str, feature_schema_id: Cuid,
                    extra: Dict[str, Any]) -> "NDSegments":

        segments = [NDSegment.from_common(segment) for segment in segments]

        return cls(segments=segments,
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=extra.get('uuid'))


class NDDicomSegments(NDBaseObject, DicomSupported):
    segments: List[NDDicomSegment]

    def to_common(self, name: str, feature_schema_id: Cuid):
        result = []
        for idx, segment in enumerate(self.segments):
            result.extend(
                segment.to_common(name=name,
                                  feature_schema_id=feature_schema_id,
                                  segment_index=idx,
                                  uuid=self.uuid,
                                  group_key=self.group_key))
        return result

    @classmethod
    def from_common(cls, segments: List[DICOMObjectAnnotation], data: VideoData,
                    name: str, feature_schema_id: Cuid, extra: Dict[str, Any],
                    group_key: str) -> "NDDicomSegments":

        segments = [NDDicomSegment.from_common(segment) for segment in segments]

        return cls(segments=segments,
                   dataRow=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=extra.get('uuid'),
                   group_key=group_key)


class _URIMask(pydantic_compat.BaseModel):
    instanceURI: str
    colorRGB: Tuple[int, int, int]


class _PNGMask(pydantic_compat.BaseModel):
    png: str


class NDMask(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    mask: Union[_URIMask, _PNGMask]

    def to_common(self) -> Mask:
        if isinstance(self.mask, _URIMask):
            return Mask(mask=MaskData(url=self.mask.instanceURI),
                        color=self.mask.colorRGB)
        else:
            encoded_image_bytes = self.mask.png.encode('utf-8')
            image_bytes = base64.b64decode(encoded_image_bytes)
            image = np.array(Image.open(BytesIO(image_bytes)))
            if np.max(image) > 1:
                raise ValueError(
                    f"Expected binary mask. Found max value of {np.max(image)}")
            # Color is 1,1,1 because it is a binary array and we are just stacking it into 3 channels
            return Mask(mask=MaskData.from_2D_arr(image), color=(1, 1, 1))

    @classmethod
    def from_common(
            cls,
            uuid: str,
            mask: Mask,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None) -> "NDMask":

        if mask.mask.url is not None:
            lbv1_mask = _URIMask(instanceURI=mask.mask.url, colorRGB=mask.color)
        else:
            binary = np.all(mask.mask.value == mask.color, axis=-1)
            im_bytes = BytesIO()
            Image.fromarray(binary, 'L').save(im_bytes, format="PNG")
            lbv1_mask = _PNGMask(
                png=base64.b64encode(im_bytes.getvalue()).decode('utf-8'))

        return cls(mask=lbv1_mask,
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDVideoMasksFramesInstances(pydantic_compat.BaseModel):
    frames: List[MaskFrame]
    instances: List[MaskInstance]


class NDVideoMasks(NDJsonBase, ConfidenceMixin, CustomMetricsNotSupportedMixin):
    masks: NDVideoMasksFramesInstances

    def to_common(self) -> VideoMaskAnnotation:
        for mask_frame in self.masks.frames:
            if mask_frame.im_bytes:
                mask_frame.im_bytes = base64.b64decode(
                    mask_frame.im_bytes.encode('utf-8'))

        return VideoMaskAnnotation(
            frames=self.masks.frames,
            instances=self.masks.instances,
        )

    @classmethod
    def from_common(cls, annotation, data):
        for mask_frame in annotation.frames:
            if mask_frame.im_bytes:
                mask_frame.im_bytes = base64.b64encode(
                    mask_frame.im_bytes).decode('utf-8')

        return cls(
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            masks=NDVideoMasksFramesInstances(frames=annotation.frames,
                                              instances=annotation.instances),
        )


class NDDicomMasks(NDVideoMasks, DicomSupported):

    def to_common(self) -> DICOMMaskAnnotation:
        return DICOMMaskAnnotation(
            frames=self.masks.frames,
            instances=self.masks.instances,
            group_key=self.group_key,
        )

    @classmethod
    def from_common(cls, annotation, data):
        return cls(
            data_row=DataRow(id=data.uid, global_key=data.global_key),
            masks=NDVideoMasksFramesInstances(frames=annotation.frames,
                                              instances=annotation.instances),
            group_key=annotation.group_key.value,
        )


class Location(pydantic_compat.BaseModel):
    start: int
    end: int


class NDTextEntity(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    location: Location

    def to_common(self) -> TextEntity:
        return TextEntity(start=self.location.start, end=self.location.end)

    @classmethod
    def from_common(
            cls,
            uuid: str,
            text_entity: TextEntity,
            classifications: List[ClassificationAnnotation],
            name: str,
            feature_schema_id: Cuid,
            extra: Dict[str, Any],
            data: Union[ImageData, TextData],
            confidence: Optional[float] = None,
            custom_metrics: Optional[List[CustomMetric]] = None
    ) -> "NDTextEntity":
        return cls(location=Location(
            start=text_entity.start,
            end=text_entity.end,
        ),
                   data_row=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDDocumentEntity(NDBaseObject, ConfidenceMixin, CustomMetricsMixin):
    name: str
    text_selections: List[DocumentTextSelection]

    def to_common(self) -> DocumentEntity:
        return DocumentEntity(name=self.name,
                              text_selections=self.text_selections)

    @classmethod
    def from_common(
        cls,
        uuid: str,
        document_entity: DocumentEntity,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: Union[ImageData, TextData],
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None
    ) -> "NDDocumentEntity":

        return cls(text_selections=document_entity.text_selections,
                   dataRow=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDConversationEntity(NDTextEntity):
    message_id: str

    def to_common(self) -> ConversationEntity:
        return ConversationEntity(start=self.location.start,
                                  end=self.location.end,
                                  message_id=self.message_id)

    @classmethod
    def from_common(
        cls,
        uuid: str,
        conversation_entity: ConversationEntity,
        classifications: List[ClassificationAnnotation],
        name: str,
        feature_schema_id: Cuid,
        extra: Dict[str, Any],
        data: Union[ImageData, TextData],
        confidence: Optional[float] = None,
        custom_metrics: Optional[List[CustomMetric]] = None
    ) -> "NDConversationEntity":
        return cls(location=Location(start=conversation_entity.start,
                                     end=conversation_entity.end),
                   message_id=conversation_entity.message_id,
                   dataRow=DataRow(id=data.uid, global_key=data.global_key),
                   name=name,
                   schema_id=feature_schema_id,
                   uuid=uuid,
                   classifications=classifications,
                   confidence=confidence,
                   custom_metrics=custom_metrics)


class NDObject:

    @staticmethod
    def to_common(annotation: "NDObjectType") -> ObjectAnnotation:
        common_annotation = annotation.to_common()
        classifications = [
            NDSubclassification.to_common(annot)
            for annot in annotation.classifications
        ]
        confidence = annotation.confidence if hasattr(annotation,
                                                      'confidence') else None

        custom_metrics = annotation.custom_metrics if hasattr(
            annotation, 'custom_metrics') else None
        return ObjectAnnotation(value=common_annotation,
                                name=annotation.name,
                                feature_schema_id=annotation.schema_id,
                                classifications=classifications,
                                extra={
                                    'uuid': annotation.uuid,
                                    'page': annotation.page,
                                    'unit': annotation.unit
                                },
                                confidence=confidence,
                                custom_metrics=custom_metrics)

    @classmethod
    def from_common(
        cls,
        annotation: Union[ObjectAnnotation, List[List[VideoObjectAnnotation]],
                          VideoMaskAnnotation], data: Union[ImageData, TextData]
    ) -> Union[NDLine, NDPoint, NDPolygon, NDDocumentRectangle, NDRectangle,
               NDMask, NDTextEntity]:
        obj = cls.lookup_object(annotation)

        # if it is video segments
        if (obj == NDSegments or obj == NDDicomSegments):

            first_video_annotation = annotation[0][0]
            args = dict(
                segments=annotation,
                data=data,
                name=first_video_annotation.name,
                feature_schema_id=first_video_annotation.feature_schema_id,
                extra=first_video_annotation.extra)

            if isinstance(first_video_annotation, DICOMObjectAnnotation):
                group_key = first_video_annotation.group_key.value
                args.update(dict(group_key=group_key))

            return obj.from_common(**args)
        elif (obj == NDVideoMasks or obj == NDDicomMasks):
            return obj.from_common(annotation, data)

        subclasses = [
            NDSubclassification.from_common(annot)
            for annot in annotation.classifications
        ]
        optional_kwargs = {}
        if (annotation.confidence):
            optional_kwargs['confidence'] = annotation.confidence

        if (annotation.custom_metrics):
            optional_kwargs['custom_metrics'] = annotation.custom_metrics

        return obj.from_common(str(annotation._uuid), annotation.value,
                               subclasses, annotation.name,
                               annotation.feature_schema_id, annotation.extra,
                               data, **optional_kwargs)

    @staticmethod
    def lookup_object(
            annotation: Union[ObjectAnnotation, List]) -> "NDObjectType":

        if isinstance(annotation, DICOMMaskAnnotation):
            result = NDDicomMasks
        elif isinstance(annotation, VideoMaskAnnotation):
            result = NDVideoMasks
        elif isinstance(annotation, list):
            try:
                first_annotation = annotation[0][0]
            except IndexError:
                raise ValueError("Annotation list cannot be empty")

            if isinstance(first_annotation, DICOMObjectAnnotation):
                result = NDDicomSegments
            else:
                result = NDSegments
        else:
            result = {
                Line: NDLine,
                Point: NDPoint,
                Polygon: NDPolygon,
                Rectangle: NDRectangle,
                DocumentRectangle: NDDocumentRectangle,
                Mask: NDMask,
                TextEntity: NDTextEntity,
                DocumentEntity: NDDocumentEntity,
                ConversationEntity: NDConversationEntity,
            }.get(type(annotation.value))
        if result is None:
            raise TypeError(
                f"Unable to convert object to MAL format. `{type(annotation.value)}`"
            )
        return result


# NOTE: Deserialization of subclasses in pydantic is a known PIA, see here https://blog.devgenius.io/deserialize-child-classes-with-pydantic-that-gonna-work-784230e1cf83
# I could implement the registry approach suggested there, but I found that if I list subclass (that has more attributes) before the parent class, it works
# This is a bit of a hack, but it works for now
NDEntityType = Union[NDConversationEntity, NDTextEntity]
NDObjectType = Union[NDLine, NDPolygon, NDPoint, NDDocumentRectangle,
                     NDRectangle, NDMask, NDEntityType, NDDocumentEntity]

NDFrameObjectType = NDFrameRectangle, NDFramePoint, NDFrameLine
NDDicomObjectType = NDDicomLine



import time

import pytest
import labelbox as lb
from labelbox.data.annotation_types.data.video import VideoData
import labelbox.types as lb_types
from labelbox.schema.annotation_import import AnnotationImportState


@pytest.fixture
def user_id(client):
    return client.get_user().uid


@pytest.fixture
def org_id(client):
    return client.get_organization().uid


def test_export_v2_video(
    client,
    configured_project_without_data_rows,
    video_data,
    video_data_row,
    bbox_video_annotation_objects,
    rand_gen,
):

    project = configured_project_without_data_rows
    project_id = project.uid
    labels = []

    _, data_row_uids = video_data
    project.create_batch(
        rand_gen(str),
        data_row_uids,  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )

    for data_row_uid in data_row_uids:
        labels = [
            lb_types.Label(data=VideoData(uid=data_row_uid),
                           annotations=bbox_video_annotation_objects)
        ]

    label_import = lb.LabelImport.create_from_objects(
        client, project_id, f'test-import-{project_id}', labels)
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.errors) == 0

    num_retries = 5
    task = None

    while (num_retries > 0):
        task = project.export_v2(
            params={
                "performance_details": False,
                "label_details": True,
                "interpolated_frames": True
            })
        task.wait_till_done()
        assert task.status == "COMPLETE"
        assert task.errors is None
        if len(task.result) == 0:
            num_retries -= 1
            time.sleep(5)
        else:
            break

    export_data = task.result
    data_row_export = export_data[0]['data_row']
    assert data_row_export['global_key'] == video_data_row['global_key']
    assert data_row_export['row_data'] == video_data_row['row_data']
    assert export_data[0]['media_attributes']['mime_type'] == 'video/mp4'
    assert export_data[0]['media_attributes'][
        'frame_rate'] == 10  # as per the video_data fixture
    assert export_data[0]['media_attributes'][
        'frame_count'] == 100  # as per the video_data fixture
    expected_export_label = {
        'label_kind': 'Video',
        'version': '1.0.0',
        'id': 'clgjnpysl000xi3zxtnp29fug',
        'label_details': {
            'created_at': '2023-04-16T17:04:23+00:00',
            'updated_at': '2023-04-16T17:04:23+00:00',
            'created_by': 'vbrodsky@labelbox.com',
            'content_last_updated_at': '2023-04-16T17:04:23+00:00',
            'reviews': []
        },
        'annotations': {
            'frames': {
                '13': {
                    'objects': {
                        'clgjnpyse000ui3zx6fr1d880': {
                            'feature_id': 'clgjnpyse000ui3zx6fr1d880',
                            'name': 'bbox',
                            'annotation_kind': 'VideoBoundingBox',
                            'classifications': [{
                                'feature_id': 'clgjnpyse000vi3zxtgtfh01y',
                                'name': 'nested',
                                'radio_answer': {
                                    'feature_id': 'clgjnpyse000wi3zxnxgv53ps',
                                    'name': 'radio_option_1',
                                    'classifications': []
                                }
                            }],
                            'bounding_box': {
                                'top': 98.0,
                                'left': 146.0,
                                'height': 243.0,
                                'width': 236.0
                            }
                        }
                    },
                    'classifications': []
                },
                '18': {
                    'objects': {
                        'clgjnpyse000ui3zx6fr1d880': {
                            'feature_id': 'clgjnpyse000ui3zx6fr1d880',
                            'name': 'bbox',
                            'annotation_kind': 'VideoBoundingBox',
                            'classifications': [{
                                'feature_id': 'clgjnpyse000vi3zxtgtfh01y',
                                'name': 'nested',
                                'radio_answer': {
                                    'feature_id': 'clgjnpyse000wi3zxnxgv53ps',
                                    'name': 'radio_option_1',
                                    'classifications': []
                                }
                            }],
                            'bounding_box': {
                                'top': 98.0,
                                'left': 146.0,
                                'height': 243.0,
                                'width': 236.0
                            }
                        }
                    },
                    'classifications': []
                },
                '19': {
                    'objects': {
                        'clgjnpyse000ui3zx6fr1d880': {
                            'feature_id': 'clgjnpyse000ui3zx6fr1d880',
                            'name': 'bbox',
                            'annotation_kind': 'VideoBoundingBox',
                            'classifications': [],
                            'bounding_box': {
                                'top': 98.0,
                                'left': 146.0,
                                'height': 243.0,
                                'width': 236.0
                            }
                        }
                    },
                    'classifications': []
                }
            },
            'segments': {
                'clgjnpyse000ui3zx6fr1d880': [[13, 13], [18, 19]]
            },
            'key_frame_feature_map': {
                'clgjnpyse000ui3zx6fr1d880': {
                    '13': True,
                    '18': False,
                    '19': True
                }
            },
            'classifications': []
        }
    }

    project_export_labels = export_data[0]['projects'][project_id]['labels']
    assert (len(project_export_labels) == len(labels)
           )  #note we create 1 label per data row, 1 data row so 1 label
    export_label = project_export_labels[0]
    assert (export_label['label_kind']) == 'Video'

    assert (export_label['label_details'].keys()
           ) == expected_export_label['label_details'].keys()

    expected_frames_ids = [
        vannotation.frame for vannotation in bbox_video_annotation_objects
    ]
    export_annotations = export_label['annotations']
    export_frames = export_annotations['frames']
    export_frames_ids = [int(frame_id) for frame_id in export_frames.keys()]
    all_frames_exported = []
    for value in expected_frames_ids:  # note need to understand why we are exporting more frames than we created
        if value not in export_frames_ids:
            all_frames_exported.append(value)
    assert (len(all_frames_exported) == 0)

    # BEGINNING OF THE VIDEO INTERPOLATION ASSERTIONS
    first_frame_id = bbox_video_annotation_objects[0].frame
    last_frame_id = bbox_video_annotation_objects[-1].frame

    # Generate list of frames with frames in between, e.g. 13, 14, 15, 16, 17, 18, 19
    expected_frame_ids = list(range(first_frame_id, last_frame_id + 1))

    assert export_frames_ids == expected_frame_ids

    exported_objects_dict = export_frames[str(first_frame_id)]['objects']

    # Get the label ID
    first_exported_label_id = list(exported_objects_dict.keys())[0]

    # Since the bounding box moves to the right, the interpolated frame content should start a little bit more far to the right
    assert export_frames[str(first_frame_id + 1)]['objects'][
        first_exported_label_id]['bounding_box']['left'] > export_frames[
            str(first_frame_id
               )]['objects'][first_exported_label_id]['bounding_box']['left']
    # But it shouldn't be further than the last frame
    assert export_frames[str(first_frame_id + 1)]['objects'][
        first_exported_label_id]['bounding_box']['left'] < export_frames[
            str(last_frame_id
               )]['objects'][first_exported_label_id]['bounding_box']['left']
    # END OF THE VIDEO INTERPOLATION ASSERTIONS

    frame_with_nested_classifications = export_frames['13']
    annotation = None
    for _, a in frame_with_nested_classifications['objects'].items():
        if a['name'] == 'bbox':
            annotation = a
            break
    assert (annotation is not None)
    assert (annotation['annotation_kind'] == 'VideoBoundingBox')
    assert (annotation['classifications'])
    assert (annotation['bounding_box'] == {
        'top': 98.0,
        'left': 146.0,
        'height': 243.0,
        'width': 236.0
    })
    classifications = annotation['classifications']
    classification = classifications[0]['radio_answer']
    assert (classification['name'] == 'radio_option_1')
    subclassifications = classification['classifications']
    # NOTE predictions services does not support nested classifications at the moment, see
    # https://labelbox.atlassian.net/browse/AL-5588
    assert (len(subclassifications) == 0)



import pytest
from labelbox.exceptions import ResourceNotFoundError


def test_add_single_model_config(live_chat_evaluation_project_with_new_dataset,
                                 model_config):
    configured_project = live_chat_evaluation_project_with_new_dataset
    project_model_config_id = configured_project.add_model_config(
        model_config.uid)

    assert set(config.uid
               for config in configured_project.project_model_configs()) == set(
                   [project_model_config_id])

    assert configured_project.delete_project_model_config(
        project_model_config_id)


def test_add_multiple_model_config(client, rand_gen,
                                   live_chat_evaluation_project_with_new_dataset,
                                   model_config, valid_model_id):
    configured_project = live_chat_evaluation_project_with_new_dataset
    second_model_config = client.create_model_config(rand_gen(str),
                                                     valid_model_id,
                                                     {"param": "value"})
    first_project_model_config_id = configured_project.add_model_config(
        model_config.uid)
    second_project_model_config_id = configured_project.add_model_config(
        second_model_config.uid)
    expected_model_configs = set(
        [first_project_model_config_id, second_project_model_config_id])

    assert set(
        config.uid for config in configured_project.project_model_configs()
    ) == expected_model_configs

    for project_model_config_id in expected_model_configs:
        assert configured_project.delete_project_model_config(
            project_model_config_id)


def test_delete_project_model_config(live_chat_evaluation_project_with_new_dataset,
                                     model_config):
    configured_project = live_chat_evaluation_project_with_new_dataset
    assert configured_project.delete_project_model_config(
        configured_project.add_model_config(model_config.uid))
    assert not len(configured_project.project_model_configs())


def test_delete_nonexistant_project_model_config(configured_project):
    with pytest.raises(ResourceNotFoundError):
        configured_project.delete_project_model_config(
            "nonexistant_project_model_config")



from labelbox.data.annotation_types.annotation import ClassificationAnnotation
from labelbox.data.annotation_types.classification.classification import ClassificationAnswer, Radio, Text
from labelbox.data.annotation_types.data.text import TextData
from labelbox.data.annotation_types.label import Label

from labelbox.data.serialization.ndjson.converter import NDJsonConverter


def test_serialization():
    label = Label(uid="ckj7z2q0b0000jx6x0q2q7q0d",
                  data=TextData(
                      uid="bkj7z2q0b0000jx6x0q2q7q0d",
                      text="This is a test",
                  ),
                  annotations=[
                      ClassificationAnnotation(
                          name="radio_question_geo",
                          confidence=0.5,
                          value=Text(answer="first_radio_answer"))
                  ])

    serialized = NDJsonConverter.serialize([label])
    res = next(serialized)
    assert 'confidence' not in res  # because confidence needs to be set on the annotation itself
    assert res['name'] == "radio_question_geo"
    assert res['answer'] == "first_radio_answer"
    assert res['dataRow']['id'] == "bkj7z2q0b0000jx6x0q2q7q0d"

    deserialized = NDJsonConverter.deserialize([res])
    res = next(deserialized)
    annotation = res.annotations[0]

    annotation_value = annotation.value
    assert type(annotation_value) is Text
    assert annotation_value.answer == "first_radio_answer"



from typing import Callable, Optional, Tuple, Union, Dict, List

import numpy as np
import cv2

from shapely.geometry import MultiPolygon, Polygon

from ..data import MaskData
from .geometry import Geometry

from labelbox import pydantic_compat


class Mask(Geometry):
    """Mask used to represent a single class in a larger segmentation mask

    Example of a mutually exclusive class

    >>> arr = MaskData.from_2D_arr([
    >>>    [0, 0, 0],
    >>>    [1, 1, 1],
    >>>    [2, 2, 2],
    >>>])
    >>> annotations = [
    >>>    ObjectAnnotation(value=Mask(mask=arr, color=1), name="dog"),
    >>>    ObjectAnnotation(value=Mask(mask=arr, color=2), name="cat"),
    >>>]

    Args:
         mask (MaskData): An object containing the actual mask, `MaskData` can
            be shared across multiple `Masks` to more efficiently store data
            for mutually exclusive segmentations.
         color (Tuple[uint8, uint8, uint8]): RGB color or a single value
            indicating the values of the class in the `MaskData`
    """

    mask: MaskData
    color: Union[Tuple[int, int, int], int]

    @property
    def geometry(self) -> Dict[str, Tuple[int, int, int]]:
        mask = self.draw(color=1)
        contours, hierarchy = cv2.findContours(image=mask,
                                               mode=cv2.RETR_TREE,
                                               method=cv2.CHAIN_APPROX_NONE)

        holes = []
        external_contours = []
        for i in range(len(contours)):
            if hierarchy[0, i, 3] != -1:
                #determined to be a hole based on contour hierarchy
                holes.append(contours[i])
            else:
                external_contours.append(contours[i])

        external_polygons = self._extract_polygons_from_contours(
            external_contours)
        holes = self._extract_polygons_from_contours(holes)

        if not external_polygons.is_valid:
            external_polygons = external_polygons.buffer(0)

        if not holes.is_valid:
            holes = holes.buffer(0)

        return external_polygons.difference(holes).__geo_interface__

    def draw(self,
             height: Optional[int] = None,
             width: Optional[int] = None,
             canvas: Optional[np.ndarray] = None,
             color: Optional[Union[int, Tuple[int, int, int]]] = None,
             thickness=None) -> np.ndarray:
        """Converts the Mask object into a numpy array

        Args:
            height (int): Optionally resize mask height before drawing.
            width (int): Optionally resize mask width before drawing.
            canvas (np.ndarray): Optionall provide a canvas to draw on
            color (Union[int, Tuple[int,int,int]]): Color to draw the canvas.
                Defaults to using the encoded color in the mask.
                int will return the mask as a 1d array
                tuple[int,int,int] will return the mask as a 3d array
            thickness (None): Unused, exists for a consistent interface.

        Returns:
            np.ndarray representing only this object
                as opposed to the mask that this object references which might have multiple objects determined by colors
        """
        mask = self.mask.value
        mask = np.alltrue(mask == self.color, axis=2).astype(np.uint8)

        if height is not None or width is not None:
            mask = cv2.resize(mask,
                              (width or mask.shape[1], height or mask.shape[0]))

        dims = [mask.shape[0], mask.shape[1]]
        color = color or self.color
        if isinstance(color, (tuple, list)):
            dims = dims + [len(color)]

        canvas = canvas if canvas is not None else np.zeros(tuple(dims),
                                                            dtype=np.uint8)
        canvas[mask.astype(bool)] = color
        return canvas

    def _extract_polygons_from_contours(self, contours: List) -> MultiPolygon:
        contours = map(np.squeeze, contours)
        filtered_contours = filter(lambda contour: len(contour) > 2, contours)
        polygons = map(Polygon, filtered_contours)
        return MultiPolygon(polygons)

    def create_url(self, signer: Callable[[bytes], str]) -> str:
        """
        Update the segmentation mask to have a url.
        Only update the mask if it doesn't already have a url

        Args:
            signer: A function that accepts bytes and returns a signed url.
        Returns:
            the url for the mask
        """
        return self.mask.create_url(signer)

    @pydantic_compat.validator('color')
    def is_valid_color(cls, color):
        if isinstance(color, (tuple, list)):
            if len(color) == 1:
                color = [color[0]] * 3
            if len(color) != 3:
                raise ValueError(
                    "Segmentation colors must be either a (r,g,b) tuple or a single grayscale value"
                )
            elif not all([0 <= c <= 255 for c in color]):
                raise ValueError(
                    f"All rgb colors must be between 0 and 255. Found : {color}"
                )
        elif not (0 <= color <= 255):
            raise ValueError(
                f"All rgb colors must be between 0 and 255. Found : {color}")

        return color



import pytest

from labelbox.data.annotation_types.classification.classification import Checklist, ClassificationAnnotation, ClassificationAnswer, Radio
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.video import VideoObjectAnnotation


@pytest.fixture
def bbox_video_annotation_objects():
    bbox_annotation = [
        VideoObjectAnnotation(
            name="bbox",
            keyframe=True,
            frame=13,
            segment_index=0,
            value=Rectangle(
                start=Point(x=146.0, y=98.0),  # Top left
                end=Point(x=382.0, y=341.0),  # Bottom right
            ),
            classifications=[
                ClassificationAnnotation(
                    name='nested',
                    value=Radio(answer=ClassificationAnswer(
                        name='radio_option_1',
                        classifications=[
                            ClassificationAnnotation(
                                name='nested_checkbox',
                                value=Checklist(answer=[
                                    ClassificationAnswer(
                                        name='nested_checkbox_option_1'),
                                    ClassificationAnswer(
                                        name='nested_checkbox_option_2')
                                ]))
                        ])),
                )
            ]),
        VideoObjectAnnotation(
            name="bbox",
            keyframe=True,
            frame=19,
            segment_index=0,
            value=Rectangle(
                start=Point(x=186.0, y=98.0),  # Top left
                end=Point(x=490.0, y=341.0),  # Bottom right
            ))
    ]

    return bbox_annotation



import json
import labelbox.types as lb_types
from labelbox.data.serialization.ndjson.converter import NDJsonConverter

DATAROW_ID = "ckrb1sf1i1g7i0ybcdc6oc8ct"


def test_rectangle():
    with open('tests/data/assets/ndjson/rectangle_import.json', 'r') as file:
        data = json.load(file)
    res = list(NDJsonConverter.deserialize(data))
    res = list(NDJsonConverter.serialize(res))
    assert res == data


def test_rectangle_inverted_start_end_points():
    with open('tests/data/assets/ndjson/rectangle_import.json', 'r') as file:
        data = json.load(file)

    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=81, y=69),
            end=lb_types.Point(x=38, y=28),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"})

    label = lb_types.Label(data={"uid":DATAROW_ID},
                           annotations=[bbox])

    res = list(NDJsonConverter.serialize([label]))
    assert res == data

    expected_bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=38, y=28),
            end=lb_types.Point(x=81, y=69),
        ),
        extra={
            "uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72",
            "page": None,
            "unit": None
        })

    label = lb_types.Label(data={"uid":DATAROW_ID},
                           annotations=[expected_bbox])

    res = list(NDJsonConverter.deserialize(res))
    assert res == [label]


def test_rectangle_mixed_start_end_points():
    with open('tests/data/assets/ndjson/rectangle_import.json', 'r') as file:
        data = json.load(file)

    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=81, y=28),
            end=lb_types.Point(x=38, y=69),
        ),
        extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"})

    label = lb_types.Label(data={"uid":DATAROW_ID},
                           annotations=[bbox])

    res = list(NDJsonConverter.serialize([label]))
    assert res == data

    bbox = lb_types.ObjectAnnotation(
        name="bbox",
        value=lb_types.Rectangle(
            start=lb_types.Point(x=38, y=28),
            end=lb_types.Point(x=81, y=69),
        ),
        extra={
            "uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72",
            "page": None,
            "unit": None
        })

    label = lb_types.Label(data={"uid":DATAROW_ID},
                           annotations=[bbox])

    res = list(NDJsonConverter.deserialize(res))
    assert res == [label]



from enum import Enum


class BulkImportRequestState(Enum):
    """ State of the import job when importing annotations (RUNNING, FAILED, or FINISHED).

       If you are not usinig MEA continue using BulkImportRequest.
       AnnotationImports are in beta and will change soon.

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - RUNNING
         - Indicates that the import job is not done yet.
       * - FAILED
         - Indicates the import job failed. Check `BulkImportRequest.errors` for more information
       * - FINISHED
         - Indicates the import job is no longer running. Check `BulkImportRequest.statuses` for more information
    """
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    FINISHED = "FINISHED"


class AnnotationImportState(Enum):
    """ State of the import job when importing annotations (RUNNING, FAILED, or FINISHED).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - RUNNING
         - Indicates that the import job is not done yet.
       * - FAILED
         - Indicates the import job failed. Check `AnnotationImport.errors` for more information
       * - FINISHED
         - Indicates the import job is no longer running. Check `AnnotationImport.statuses` for more information
    """
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    FINISHED = "FINISHED"


class CollectionJobStatus(Enum):
    """ Status of an asynchronous job over a collection.

        * - State
          - Description
        * - SUCCESS
          - Indicates job has successfully processed entire collection of data
        * - PARTIAL SUCCESS
          - Indicates some data in the collection has succeeded and other data have failed
        * - FAILURE
          - Indicates job has failed to process entire collection of data
    """
    SUCCESS = "SUCCESS"
    PARTIAL_SUCCESS = "PARTIAL SUCCESS"
    FAILURE = "FAILURE"


import sys
from hashlib import md5

from .... import pydantic_compat


class Categories(pydantic_compat.BaseModel):
    id: int
    name: str
    supercategory: str
    isthing: int = 1


def hash_category_name(name: str) -> int:
    return int.from_bytes(
        md5(name.encode('utf-8')).hexdigest().encode('utf-8'), 'little')



import datetime
from labelbox.data.annotation_types.data.generic_data_row_data import GenericDataRowData
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.data.annotation_types import Label
import pytest
import uuid

import labelbox as lb
from labelbox.schema.media_type import MediaType
from labelbox.schema.annotation_import import AnnotationImportState
from labelbox import Project, Client
import itertools

"""
 - integration test for importing mal labels and ground truths with each supported MediaType. 
 - NDJSON is used to generate annotations.
"""

def validate_iso_format(date_string: str):
    parsed_t = datetime.datetime.fromisoformat(
        date_string)  # this will blow up if the string is not in iso format
    assert parsed_t.hour is not None
    assert parsed_t.minute is not None
    assert parsed_t.second is not None
    
@pytest.mark.parametrize(
    "media_type, data_type_class",
    [
        (MediaType.Audio, GenericDataRowData),
        (MediaType.Html, GenericDataRowData),
        (MediaType.Image, GenericDataRowData),
        (MediaType.Text, GenericDataRowData),
        (MediaType.Video, GenericDataRowData),
        (MediaType.Conversational, GenericDataRowData),
        (MediaType.Document, GenericDataRowData),
    ],
)
def test_generic_data_row_type_by_data_row_id(
    media_type,
    data_type_class,
    annotations_by_media_type,
    hardcoded_datarow_id,
):
    annotations_ndjson =  annotations_by_media_type[media_type]
    annotations_ndjson = [annotation[0] for annotation in annotations_ndjson]
    
    label = list(NDJsonConverter.deserialize(annotations_ndjson))[0]
   
    data_label = Label(data=data_type_class(uid = hardcoded_datarow_id()),
                       annotations=label.annotations)

    assert data_label.data.uid == label.data.uid
    assert label.annotations == data_label.annotations


@pytest.mark.parametrize(
    "media_type, data_type_class",
    [
        (MediaType.Audio, GenericDataRowData),
        (MediaType.Html, GenericDataRowData),
        (MediaType.Image, GenericDataRowData),
        (MediaType.Text, GenericDataRowData),
        (MediaType.Video, GenericDataRowData),
        (MediaType.Conversational, GenericDataRowData),
        (MediaType.Document, GenericDataRowData),
    ],
)
def test_generic_data_row_type_by_global_key(
    media_type,
    data_type_class,
    annotations_by_media_type,
    hardcoded_global_key,
):
    annotations_ndjson =  annotations_by_media_type[media_type]
    annotations_ndjson = [annotation[0] for annotation in annotations_ndjson]
    
    label = list(NDJsonConverter.deserialize(annotations_ndjson))[0]
   
    data_label = Label(data=data_type_class(global_key = hardcoded_global_key()),
                       annotations=label.annotations)

    assert data_label.data.global_key == label.data.global_key
    assert label.annotations == data_label.annotations


# TODO: add MediaType.LLMPromptResponseCreation(data gen) once supported and llm human preference once media type is added
@pytest.mark.parametrize(
    "configured_project",
    [
        MediaType.Audio,
        MediaType.Html,
        MediaType.Image,
        MediaType.Text,
        MediaType.Video,
        MediaType.Conversational,
        MediaType.Document,
        MediaType.Dicom,
    ],
    indirect=True
)
def test_import_media_types(
    client: Client,
    configured_project: Project,
    annotations_by_media_type,
    exports_v2_by_media_type,
    export_v2_test_helpers,
    helpers,
):
    annotations_ndjson =  list(itertools.chain.from_iterable(annotations_by_media_type[configured_project.media_type]))

    label_import = lb.LabelImport.create_from_objects(
        client, configured_project.uid, f"test-import-{configured_project.media_type}", annotations_ndjson)
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.errors) == 0

    result = export_v2_test_helpers.run_project_export_v2_task(configured_project)

    assert result

    for exported_data in result:
        # timestamp fields are in iso format
        validate_iso_format(exported_data["data_row"]["details"]["created_at"])
        validate_iso_format(exported_data["data_row"]["details"]["updated_at"])
        validate_iso_format(exported_data["projects"][configured_project.uid]["labels"][0]
                            ["label_details"]["created_at"])
        validate_iso_format(exported_data["projects"][configured_project.uid]["labels"][0]
                            ["label_details"]["updated_at"])

        assert exported_data["data_row"]["id"] in configured_project.data_row_ids
        exported_project = exported_data["projects"][configured_project.uid]
        exported_project_labels = exported_project["labels"][0]
        exported_annotations = exported_project_labels["annotations"]

        expected_data = exports_v2_by_media_type[configured_project.media_type]
        helpers.remove_keys_recursive(exported_annotations,
                                    ["feature_id", "feature_schema_id"])     
        helpers.rename_cuid_key_recursive(exported_annotations)

        assert exported_annotations == expected_data 


@pytest.mark.order(1)
@pytest.mark.parametrize(
    "configured_project_by_global_key",
    [
        MediaType.Audio,
        MediaType.Html,
        MediaType.Image,
        MediaType.Text,
        MediaType.Video,
        MediaType.Conversational,
        MediaType.Document,
        MediaType.Dicom,
    ],
    indirect=True
)
def test_import_media_types_by_global_key(
    client,
    configured_project_by_global_key,
    annotations_by_media_type,
    exports_v2_by_media_type,
    export_v2_test_helpers,
    helpers,
):
    annotations_ndjson =  list(itertools.chain.from_iterable(annotations_by_media_type[configured_project_by_global_key.media_type]))

    label_import = lb.LabelImport.create_from_objects(
        client, configured_project_by_global_key.uid, f"test-import-{configured_project_by_global_key.media_type}", annotations_ndjson)
    label_import.wait_until_done()

    assert label_import.state == AnnotationImportState.FINISHED
    assert len(label_import.errors) == 0

    result = export_v2_test_helpers.run_project_export_v2_task(configured_project_by_global_key)

    assert result

    for exported_data in result:
        # timestamp fields are in iso format
        validate_iso_format(exported_data["data_row"]["details"]["created_at"])
        validate_iso_format(exported_data["data_row"]["details"]["updated_at"])
        validate_iso_format(exported_data["projects"][configured_project_by_global_key.uid]["labels"][0]
                            ["label_details"]["created_at"])
        validate_iso_format(exported_data["projects"][configured_project_by_global_key.uid]["labels"][0]
                            ["label_details"]["updated_at"])

        assert exported_data["data_row"]["id"] in configured_project_by_global_key.data_row_ids
        exported_project = exported_data["projects"][configured_project_by_global_key.uid]
        exported_project_labels = exported_project["labels"][0]
        exported_annotations = exported_project_labels["annotations"]

        expected_data = exports_v2_by_media_type[configured_project_by_global_key.media_type]
        helpers.remove_keys_recursive(exported_annotations,
                                    ["feature_id", "feature_schema_id"])     
        helpers.rename_cuid_key_recursive(exported_annotations)

        assert exported_annotations == expected_data 


@pytest.mark.parametrize(
    "configured_project",
    [
        MediaType.Audio,
        MediaType.Html,
        MediaType.Image,
        MediaType.Text,
        MediaType.Video,
        MediaType.Conversational,
        MediaType.Document,
        MediaType.Dicom,
    ],
    indirect=True
)
def test_import_mal_annotations(
    client,
    configured_project: Project,
    annotations_by_media_type,
):
    annotations_ndjson =  list(itertools.chain.from_iterable(annotations_by_media_type[configured_project.media_type]))

    import_annotations = lb.MALPredictionImport.create_from_objects(
        client=client,
        project_id=configured_project.uid,
        name=f"import {str(uuid.uuid4())}",
        predictions=annotations_ndjson,
    )
    import_annotations.wait_until_done()

    assert import_annotations.errors == []
    # MAL Labels cannot be exported and compared to input labels
    

@pytest.mark.parametrize(
    "configured_project_by_global_key",
    [
        MediaType.Audio,
        MediaType.Html,
        MediaType.Image,
        MediaType.Text,
        MediaType.Video,
        MediaType.Conversational,
        MediaType.Document,
        MediaType.Dicom,
    ],
    indirect=True
)
def test_import_mal_annotations_global_key(client,
                                           configured_project_by_global_key: Project,
                                           annotations_by_media_type):

    annotations_ndjson =  list(itertools.chain.from_iterable(annotations_by_media_type[configured_project_by_global_key.media_type]))

    import_annotations = lb.MALPredictionImport.create_from_objects(
        client=client,
        project_id=configured_project_by_global_key.uid,
        name=f"import {str(uuid.uuid4())}",
        predictions=annotations_ndjson,
    )
    import_annotations.wait_until_done()

    assert import_annotations.errors == []
    # MAL Labels cannot be exported and compared to input labels



from labelbox.data.serialization.labelbox_v1.objects import LBV1Mask
from typing import Any, Dict, Generator, Iterable, Union
import logging

from labelbox import parser
import requests
from copy import deepcopy
from requests.exceptions import HTTPError
from google.api_core import retry

import labelbox
from .label import LBV1Label
from ...annotation_types.collection import (LabelCollection, LabelGenerator,
                                            PrefetchGenerator)

logger = logging.getLogger(__name__)


class LBV1Converter:

    @staticmethod
    def deserialize_video(json_data: Union[str, Iterable[Dict[str, Any]]],
                          client: "labelbox.Client") -> LabelGenerator:
        """
        Converts a labelbox video export into the common labelbox format.

        Args:
            json_data: An iterable representing the labelbox video export.
            client: The labelbox client for downloading video annotations
        Returns:
            LabelGenerator containing the video data.
        """
        label_generator = (LBV1Label(**example).to_common()
                           for example in LBV1VideoIterator(json_data, client)
                           if example['Label'])
        return LabelGenerator(data=label_generator)

    @staticmethod
    def deserialize(
            json_data: Union[str, Iterable[Dict[str, Any]]]) -> LabelGenerator:
        """
        Converts a labelbox export (non-video) into the common labelbox format.

        Args:
            json_data: An iterable representing the labelbox export.
        Returns:
            LabelGenerator containing the export data.
        """

        def label_generator():
            for example in json_data:
                if 'frames' in example['Label']:
                    raise ValueError(
                        "Use `LBV1Converter.deserialize_video` to process video"
                    )

                if example['Label']:
                    # Don't construct empty dict
                    yield LBV1Label(**example).to_common()

        return LabelGenerator(data=label_generator())

    @staticmethod
    def serialize(
            labels: LabelCollection) -> Generator[Dict[str, Any], None, None]:
        """
        Converts a labelbox common object to the labelbox json export format

        Note that any metric annotations will not be written since they are not defined in the LBV1 format.

        Args:
            labels: Either a list of Label objects or a LabelGenerator (LabelCollection)
        Returns:
            A generator for accessing the labelbox json export representation of the data
        """
        for label in labels:
            res = LBV1Label.from_common(label)
            yield res.dict(by_alias=True)


class LBV1VideoIterator(PrefetchGenerator):
    """
    Generator that fetches video annotations in the background to be faster.
    """

    def __init__(self, examples, client):
        self.client = client
        super().__init__(examples)

    def _process(self, value):
        value = deepcopy(value)
        if 'frames' in value['Label']:
            req = self._request(value)
            value['Label'] = parser.loads(req)
        return value

    @retry.Retry(predicate=retry.if_exception_type(HTTPError))
    def _request(self, value):
        req = requests.get(
            value['Label']['frames'],
            headers={"Authorization": f"Bearer {self.client.api_key}"})
        if req.status_code == 401:
            raise labelbox.exceptions.AuthenticationError("Invalid API key")
        req.raise_for_status()
        return req.text



import json
import time
from uuid import UUID, uuid4
import functools

import logging
from pathlib import Path
from google.api_core import retry
from labelbox import parser
import requests
from labelbox import pydantic_compat
from typing_extensions import Literal
from typing import (Any, List, Optional, BinaryIO, Dict, Iterable, Tuple, Union,
                    Type, Set, TYPE_CHECKING)

from labelbox import exceptions as lb_exceptions
from labelbox.orm.model import Entity
from labelbox import utils
from labelbox.orm import query
from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field, Relationship
from labelbox.schema.enums import BulkImportRequestState
from labelbox.schema.serialization import serialize_labels

if TYPE_CHECKING:
    from labelbox import Project
    from labelbox.types import Label

NDJSON_MIME_TYPE = "application/x-ndjson"
logger = logging.getLogger(__name__)


def _make_file_name(project_id: str, name: str) -> str:
    return f"{project_id}__{name}.ndjson"


# TODO(gszpak): move it to client.py
def _make_request_data(project_id: str, name: str, content_length: int,
                       file_name: str) -> dict:
    query_str = """mutation createBulkImportRequestFromFilePyApi(
            $projectId: ID!, $name: String!, $file: Upload!, $contentLength: Int!) {
        createBulkImportRequest(data: {
            projectId: $projectId,
            name: $name,
            filePayload: {
                file: $file,
                contentLength: $contentLength
            }
        }) {
            %s
        }
    }
    """ % query.results_query_part(BulkImportRequest)
    variables = {
        "projectId": project_id,
        "name": name,
        "file": None,
        "contentLength": content_length
    }
    operations = json.dumps({"variables": variables, "query": query_str})

    return {
        "operations": operations,
        "map": (None, json.dumps({file_name: ["variables.file"]}))
    }


def _send_create_file_command(
        client, request_data: dict, file_name: str,
        file_data: Tuple[str, Union[bytes, BinaryIO], str]) -> dict:

    response = client.execute(data=request_data, files={file_name: file_data})

    if not response.get("createBulkImportRequest", None):
        raise lb_exceptions.LabelboxError(
            "Failed to create BulkImportRequest, message: %s" %
            response.get("errors", None) or response.get("error", None))

    return response


class BulkImportRequest(DbObject):
    """Represents the import job when importing annotations.

    Attributes:
        name (str)
        state (Enum): FAILED, RUNNING, or FINISHED (Refers to the whole import job)
        input_file_url (str): URL to your web-hosted NDJSON file
        error_file_url (str): NDJSON that contains error messages for failed annotations
        status_file_url (str): NDJSON that contains status for each annotation
        created_at (datetime): UTC timestamp for date BulkImportRequest was created

        project (Relationship): `ToOne` relationship to Project
        created_by (Relationship): `ToOne` relationship to User
    """
    name = Field.String("name")
    state = Field.Enum(BulkImportRequestState, "state")
    input_file_url = Field.String("input_file_url")
    error_file_url = Field.String("error_file_url")
    status_file_url = Field.String("status_file_url")
    created_at = Field.DateTime("created_at")

    project = Relationship.ToOne("Project")
    created_by = Relationship.ToOne("User", False, "created_by")

    @property
    def inputs(self) -> List[Dict[str, Any]]:
        """
        Inputs for each individual annotation uploaded.
        This should match the ndjson annotations that you have uploaded.

        Returns:
            Uploaded ndjson.

        * This information will expire after 24 hours.
        """
        return self._fetch_remote_ndjson(self.input_file_url)

    @property
    def errors(self) -> List[Dict[str, Any]]:
        """
        Errors for each individual annotation uploaded. This is a subset of statuses

        Returns:
            List of dicts containing error messages. Empty list means there were no errors
            See `BulkImportRequest.statuses` for more details.

        * This information will expire after 24 hours.
        """
        self.wait_until_done()
        return self._fetch_remote_ndjson(self.error_file_url)

    @property
    def statuses(self) -> List[Dict[str, Any]]:
        """
        Status for each individual annotation uploaded.

        Returns:
            A status for each annotation if the upload is done running.
            See below table for more details

        .. list-table::
           :widths: 15 150
           :header-rows: 1

           * - Field
             - Description
           * - uuid
             - Specifies the annotation for the status row.
           * - dataRow
             - JSON object containing the Labelbox data row ID for the annotation.
           * - status
             - Indicates SUCCESS or FAILURE.
           * - errors
             - An array of error messages included when status is FAILURE. Each error has a name, message and optional (key might not exist) additional_info.

        * This information will expire after 24 hours.
        """
        self.wait_until_done()
        return self._fetch_remote_ndjson(self.status_file_url)

    @functools.lru_cache()
    def _fetch_remote_ndjson(self, url: str) -> List[Dict[str, Any]]:
        """
        Fetches the remote ndjson file and caches the results.

        Args:
            url (str): Can be any url pointing to an ndjson file.
        Returns:
            ndjson as a list of dicts.
        """
        response = requests.get(url)
        response.raise_for_status()
        return parser.loads(response.text)

    def refresh(self) -> None:
        """Synchronizes values of all fields with the database.
        """
        query_str, params = query.get_single(BulkImportRequest, self.uid)
        res = self.client.execute(query_str, params)
        res = res[utils.camel_case(BulkImportRequest.type_name())]
        self._set_field_values(res)

    def wait_till_done(self, sleep_time_seconds: int = 5) -> None:
        self.wait_until_done(sleep_time_seconds)

    def wait_until_done(self, sleep_time_seconds: int = 5) -> None:
        """Blocks import job until certain conditions are met.

        Blocks until the BulkImportRequest.state changes either to
        `BulkImportRequestState.FINISHED` or `BulkImportRequestState.FAILED`,
        periodically refreshing object's state.

        Args:
            sleep_time_seconds (str): a time to block between subsequent API calls
        """
        while self.state == BulkImportRequestState.RUNNING:
            logger.info(f"Sleeping for {sleep_time_seconds} seconds...")
            time.sleep(sleep_time_seconds)
            self.__exponential_backoff_refresh()

    @retry.Retry(predicate=retry.if_exception_type(lb_exceptions.ApiLimitError,
                                                   lb_exceptions.TimeoutError,
                                                   lb_exceptions.NetworkError))
    def __exponential_backoff_refresh(self) -> None:
        self.refresh()

    @classmethod
    def from_name(cls, client, project_id: str,
                  name: str) -> 'BulkImportRequest':
        """ Fetches existing BulkImportRequest.

        Args:
            client (Client): a Labelbox client
            project_id (str): BulkImportRequest's project id
            name (str): name of BulkImportRequest
        Returns:
            BulkImportRequest object

        """
        query_str = """query getBulkImportRequestPyApi(
                $projectId: ID!, $name: String!) {
            bulkImportRequest(where: {
                projectId: $projectId,
                name: $name
            }) {
                %s
            }
        }
        """ % query.results_query_part(cls)
        params = {"projectId": project_id, "name": name}
        response = client.execute(query_str, params=params)
        return cls(client, response['bulkImportRequest'])

    @classmethod
    def create_from_url(cls,
                        client,
                        project_id: str,
                        name: str,
                        url: str,
                        validate=True) -> 'BulkImportRequest':
        """
        Creates a BulkImportRequest from a publicly accessible URL
        to an ndjson file with predictions.

        Args:
            client (Client): a Labelbox client
            project_id (str): id of project for which predictions will be imported
            name (str): name of BulkImportRequest
            url (str): publicly accessible URL pointing to ndjson file containing predictions
            validate (bool): a flag indicating if there should be a validation
                if `url` is valid ndjson
        Returns:
            BulkImportRequest object
        """
        if validate:
            logger.warn(
                "Validation is turned on. The file will be downloaded locally and processed before uploading."
            )
            res = requests.get(url)
            data = parser.loads(res.text)
            _validate_ndjson(data, client.get_project(project_id))

        query_str = """mutation createBulkImportRequestPyApi(
                $projectId: ID!, $name: String!, $fileUrl: String!) {
            createBulkImportRequest(data: {
                projectId: $projectId,
                name: $name,
                fileUrl: $fileUrl
            }) {
                %s
            }
        }
        """ % query.results_query_part(cls)
        params = {"projectId": project_id, "name": name, "fileUrl": url}
        bulk_import_request_response = client.execute(query_str, params=params)
        return cls(client,
                   bulk_import_request_response["createBulkImportRequest"])

    @classmethod
    def create_from_objects(cls,
                            client,
                            project_id: str,
                            name: str,
                            predictions: Union[Iterable[Dict],
                                               Iterable["Label"]],
                            validate=True) -> 'BulkImportRequest':
        """
        Creates a `BulkImportRequest` from an iterable of dictionaries.

        Conforms to JSON predictions format, e.g.:
        ``{
            "uuid": "9fd9a92e-2560-4e77-81d4-b2e955800092",
            "schemaId": "ckappz7d700gn0zbocmqkwd9i",
            "dataRow": {
                "id": "ck1s02fqxm8fi0757f0e6qtdc"
            },
            "bbox": {
                "top": 48,
                "left": 58,
                "height": 865,
                "width": 1512
            }
        }``

        Args:
            client (Client): a Labelbox client
            project_id (str): id of project for which predictions will be imported
            name (str): name of BulkImportRequest
            predictions (Iterable[dict]): iterable of dictionaries representing predictions
            validate (bool): a flag indicating if there should be a validation
                if `predictions` is valid ndjson
        Returns:
            BulkImportRequest object
        """
        if not isinstance(predictions, list):
            raise TypeError(
                f"annotations must be in a form of Iterable. Found {type(predictions)}"
            )
        ndjson_predictions = serialize_labels(predictions)

        if validate:
            _validate_ndjson(ndjson_predictions, client.get_project(project_id))

        data_str = parser.dumps(ndjson_predictions)
        if not data_str:
            raise ValueError('annotations cannot be empty')

        data = data_str.encode('utf-8')
        file_name = _make_file_name(project_id, name)
        request_data = _make_request_data(project_id, name, len(data_str),
                                          file_name)
        file_data = (file_name, data, NDJSON_MIME_TYPE)
        response_data = _send_create_file_command(client,
                                                  request_data=request_data,
                                                  file_name=file_name,
                                                  file_data=file_data)

        return cls(client, response_data["createBulkImportRequest"])

    @classmethod
    def create_from_local_file(cls,
                               client,
                               project_id: str,
                               name: str,
                               file: Path,
                               validate_file=True) -> 'BulkImportRequest':
        """
        Creates a BulkImportRequest from a local ndjson file with predictions.

        Args:
            client (Client): a Labelbox client
            project_id (str): id of project for which predictions will be imported
            name (str): name of BulkImportRequest
            file (Path): local ndjson file with predictions
            validate_file (bool): a flag indicating if there should be a validation
                if `file` is a valid ndjson file
        Returns:
            BulkImportRequest object

        """
        file_name = _make_file_name(project_id, name)
        content_length = file.stat().st_size
        request_data = _make_request_data(project_id, name, content_length,
                                          file_name)

        with file.open('rb') as f:
            if validate_file:
                reader = parser.reader(f)
                # ensure that the underlying json load call is valid
                # https://github.com/rhgrant10/ndjson/blob/ff2f03c56b21f28f7271b27da35ca4a8bf9a05d0/ndjson/api.py#L53
                # by iterating through the file so we only store
                # each line in memory rather than the entire file
                try:
                    _validate_ndjson(reader, client.get_project(project_id))
                except ValueError:
                    raise ValueError(f"{file} is not a valid ndjson file")
                else:
                    f.seek(0)
            file_data = (file.name, f, NDJSON_MIME_TYPE)
            response_data = _send_create_file_command(client, request_data,
                                                      file_name, file_data)
        return cls(client, response_data["createBulkImportRequest"])

    def delete(self) -> None:
        """ Deletes the import job and also any annotations created by this import.

        Returns:
            None
        """
        id_param = "bulk_request_id"
        query_str = """mutation deleteBulkImportRequestPyApi($%s: ID!) {
            deleteBulkImportRequest(where: {id: $%s}) {
                id
                name
            }
        }""" % (id_param, id_param)
        self.client.execute(query_str, {id_param: self.uid})


def _validate_ndjson(lines: Iterable[Dict[str, Any]],
                     project: "Project") -> None:
    """
    Client side validation of an ndjson object.

    Does not guarentee that an upload will succeed for the following reasons:
        * We are not checking the data row types which will cause the following errors to slip through
            * Missing frame indices will not causes an error for videos
        * Uploaded annotations for the wrong data type will pass (Eg. entity on images)
        * We are not checking bounds of an asset (Eg. frame index, image height, text location)

    Args:
        lines (Iterable[Dict[str,Any]]): An iterable of ndjson lines
        project (Project): id of project for which predictions will be imported

    Raises:
        MALValidationError: Raise for invalid NDJson
        UuidError: Duplicate UUID in upload
    """
    feature_schemas_by_id, feature_schemas_by_name = get_mal_schemas(
        project.ontology())
    uids: Set[str] = set()
    for idx, line in enumerate(lines):
        try:
            annotation = NDAnnotation(**line)
            annotation.validate_instance(feature_schemas_by_id,
                                         feature_schemas_by_name)
            uuid = str(annotation.uuid)
            if uuid in uids:
                raise lb_exceptions.UuidError(
                    f'{uuid} already used in this import job, '
                    'must be unique for the project.')
            uids.add(uuid)
        except (pydantic_compat.ValidationError, ValueError, TypeError,
                KeyError) as e:
            raise lb_exceptions.MALValidationError(
                f"Invalid NDJson on line {idx}") from e


#The rest of this file contains objects for MAL validation
def parse_classification(tool):
    """
    Parses a classification from an ontology. Only radio, checklist, and text are supported for mal

    Args:
        tool (dict)

    Returns:
        dict
    """
    if tool['type'] in ['radio', 'checklist']:
        option_schema_ids = [r['featureSchemaId'] for r in tool['options']]
        option_names = [r['value'] for r in tool['options']]
        return {
            'tool': tool['type'],
            'featureSchemaId': tool['featureSchemaId'],
            'name': tool['name'],
            'options': [*option_schema_ids, *option_names]
        }
    elif tool['type'] == 'text':
        return {
            'tool': tool['type'],
            'name': tool['name'],
            'featureSchemaId': tool['featureSchemaId']
        }


def get_mal_schemas(ontology):
    """
    Converts a project ontology to a dict for easier lookup during ndjson validation

    Args:
        ontology (Ontology)
    Returns:
        Dict, Dict : Useful for looking up a tool from a given feature schema id or name
    """

    valid_feature_schemas_by_schema_id = {}
    valid_feature_schemas_by_name = {}
    for tool in ontology.normalized['tools']:
        classifications = [
            parse_classification(classification_tool)
            for classification_tool in tool['classifications']
        ]
        classifications_by_schema_id = {
            v['featureSchemaId']: v for v in classifications
        }
        classifications_by_name = {v['name']: v for v in classifications}
        valid_feature_schemas_by_schema_id[tool['featureSchemaId']] = {
            'tool': tool['tool'],
            'classificationsBySchemaId': classifications_by_schema_id,
            'classificationsByName': classifications_by_name,
            'name': tool['name']
        }
        valid_feature_schemas_by_name[tool['name']] = {
            'tool': tool['tool'],
            'classificationsBySchemaId': classifications_by_schema_id,
            'classificationsByName': classifications_by_name,
            'name': tool['name']
        }
    for tool in ontology.normalized['classifications']:
        valid_feature_schemas_by_schema_id[
            tool['featureSchemaId']] = parse_classification(tool)
        valid_feature_schemas_by_name[tool['name']] = parse_classification(tool)
    return valid_feature_schemas_by_schema_id, valid_feature_schemas_by_name


LabelboxID: str = pydantic_compat.Field(..., min_length=25, max_length=25)


class Bbox(pydantic_compat.BaseModel):
    top: float
    left: float
    height: float
    width: float


class Point(pydantic_compat.BaseModel):
    x: float
    y: float


class FrameLocation(pydantic_compat.BaseModel):
    end: int
    start: int


class VideoSupported(pydantic_compat.BaseModel):
    #Note that frames are only allowed as top level inferences for video
    frames: Optional[List[FrameLocation]]


#Base class for a special kind of union.
# Compatible with pydantic_compat. Improves error messages over a traditional union
class SpecialUnion:

    def __new__(cls, **kwargs):
        return cls.build(kwargs)

    @classmethod
    def __get_validators__(cls):
        yield cls.build

    @classmethod
    def get_union_types(cls):
        if not issubclass(cls, SpecialUnion):
            raise TypeError("{} must be a subclass of SpecialUnion")

        union_types = [x for x in cls.__orig_bases__ if hasattr(x, "__args__")]
        if len(union_types) < 1:
            raise TypeError(
                "Class {cls} should inherit from a union of objects to build")
        if len(union_types) > 1:
            raise TypeError(
                f"Class {cls} should inherit from exactly one union of objects to build. Found {union_types}"
            )
        return union_types[0].__args__[0].__args__

    @classmethod
    def build(cls: Any, data: Union[dict,
                                    pydantic_compat.BaseModel]) -> "NDBase":
        """
            Checks through all objects in the union to see which matches the input data.
            Args:
                data  (Union[dict, pydantic_compat.BaseModel]) : The data for constructing one of the objects in the union
            raises:
                KeyError: data does not contain the determinant fields for any of the types supported by this SpecialUnion
                pydantic_compat.ValidationError: Error while trying to construct a specific object in the union

        """
        if isinstance(data, pydantic_compat.BaseModel):
            data = data.dict()

        top_level_fields = []
        max_match = 0
        matched = None

        for type_ in cls.get_union_types():
            determinate_fields = type_.Config.determinants(type_)
            top_level_fields.append(determinate_fields)
            matches = sum([val in determinate_fields for val in data])
            if matches == len(determinate_fields) and matches > max_match:
                max_match = matches
                matched = type_

        if matched is not None:
            #These two have the exact same top level keys
            if matched in [NDRadio, NDText]:
                if isinstance(data['answer'], dict):
                    matched = NDRadio
                elif isinstance(data['answer'], str):
                    matched = NDText
                else:
                    raise TypeError(
                        f"Unexpected type for answer field. Found {data['answer']}. Expected a string or a dict"
                    )
            return matched(**data)
        else:
            raise KeyError(
                f"Invalid annotation. Must have one of the following keys : {top_level_fields}. Found {data}."
            )

    @classmethod
    def schema(cls):
        results = {'definitions': {}}
        for cl in cls.get_union_types():
            schema = cl.schema()
            results['definitions'].update(schema.pop('definitions'))
            results[cl.__name__] = schema
        return results


class DataRow(pydantic_compat.BaseModel):
    id: str


class NDFeatureSchema(pydantic_compat.BaseModel):
    schemaId: Optional[str] = None
    name: Optional[str] = None

    @pydantic_compat.root_validator
    def must_set_one(cls, values):
        if values['schemaId'] is None and values['name'] is None:
            raise ValueError(
                "Must set either schemaId or name for all feature schemas")
        return values


class NDBase(NDFeatureSchema):
    ontology_type: str
    uuid: UUID
    dataRow: DataRow

    def validate_feature_schemas(self, valid_feature_schemas_by_id,
                                 valid_feature_schemas_by_name):
        if self.name:
            if self.name not in valid_feature_schemas_by_name:
                raise ValueError(
                    f"Name {self.name} is not valid for the provided project's ontology."
                )

            if self.ontology_type != valid_feature_schemas_by_name[
                    self.name]['tool']:
                raise ValueError(
                    f"Name {self.name} does not map to the assigned tool {valid_feature_schemas_by_name[self.name]['tool']}"
                )

        if self.schemaId:
            if self.schemaId not in valid_feature_schemas_by_id:
                raise ValueError(
                    f"Schema id {self.schemaId} is not valid for the provided project's ontology."
                )

            if self.ontology_type != valid_feature_schemas_by_id[
                    self.schemaId]['tool']:
                raise ValueError(
                    f"Schema id {self.schemaId} does not map to the assigned tool {valid_feature_schemas_by_id[self.schemaId]['tool']}"
                )

    def validate_instance(self, valid_feature_schemas_by_id,
                          valid_feature_schemas_by_name):
        self.validate_feature_schemas(valid_feature_schemas_by_id,
                                      valid_feature_schemas_by_name)

    class Config:
        #Users shouldn't to add extra data to the payload
        extra = 'forbid'

        @staticmethod
        def determinants(parent_cls) -> List[str]:
            #This is a hack for better error messages
            return [
                k for k, v in parent_cls.__fields__.items()
                if 'determinant' in v.field_info.extra
            ]


###### Classifications ######


class NDText(NDBase):
    ontology_type: Literal["text"] = "text"
    answer: str = pydantic_compat.Field(determinant=True)
    #No feature schema to check


class NDChecklist(VideoSupported, NDBase):
    ontology_type: Literal["checklist"] = "checklist"
    answers: List[NDFeatureSchema] = pydantic_compat.Field(determinant=True)

    @pydantic_compat.validator('answers', pre=True)
    def validate_answers(cls, value, field):
        #constr not working with mypy.
        if not len(value):
            raise ValueError("Checklist answers should not be empty")
        return value

    def validate_feature_schemas(self, valid_feature_schemas_by_id,
                                 valid_feature_schemas_by_name):
        #Test top level feature schema for this tool
        super(NDChecklist,
              self).validate_feature_schemas(valid_feature_schemas_by_id,
                                             valid_feature_schemas_by_name)
        #Test the feature schemas provided to the answer field
        if len(set([answer.name or answer.schemaId for answer in self.answers
                   ])) != len(self.answers):
            raise ValueError(
                f"Duplicated featureSchema found for checklist {self.uuid}")
        for answer in self.answers:
            options = valid_feature_schemas_by_name[
                self.
                name]['options'] if self.name else valid_feature_schemas_by_id[
                    self.schemaId]['options']
            if answer.name not in options and answer.schemaId not in options:
                raise ValueError(
                    f"Feature schema provided to {self.ontology_type} invalid. Expected on of {options}. Found {answer}"
                )


class NDRadio(VideoSupported, NDBase):
    ontology_type: Literal["radio"] = "radio"
    answer: NDFeatureSchema = pydantic_compat.Field(determinant=True)

    def validate_feature_schemas(self, valid_feature_schemas_by_id,
                                 valid_feature_schemas_by_name):
        super(NDRadio,
              self).validate_feature_schemas(valid_feature_schemas_by_id,
                                             valid_feature_schemas_by_name)
        options = valid_feature_schemas_by_name[
            self.name]['options'] if self.name else valid_feature_schemas_by_id[
                self.schemaId]['options']
        if self.answer.name not in options and self.answer.schemaId not in options:
            raise ValueError(
                f"Feature schema provided to {self.ontology_type} invalid. Expected on of {options}. Found {self.answer.name or self.answer.schemaId}"
            )


#A union with custom construction logic to improve error messages
class NDClassification(
        SpecialUnion,
        Type[Union[  # type: ignore
            NDText, NDRadio, NDChecklist]]):
    ...


###### Tools ######


class NDBaseTool(NDBase):
    classifications: List[NDClassification] = []

    #This is indepdent of our problem
    def validate_feature_schemas(self, valid_feature_schemas_by_id,
                                 valid_feature_schemas_by_name):
        super(NDBaseTool,
              self).validate_feature_schemas(valid_feature_schemas_by_id,
                                             valid_feature_schemas_by_name)
        for classification in self.classifications:
            classification.validate_feature_schemas(
                valid_feature_schemas_by_name[
                    self.name]['classificationsBySchemaId']
                if self.name else valid_feature_schemas_by_id[self.schemaId]
                ['classificationsBySchemaId'], valid_feature_schemas_by_name[
                    self.name]['classificationsByName']
                if self.name else valid_feature_schemas_by_id[
                    self.schemaId]['classificationsByName'])

    @pydantic_compat.validator('classifications', pre=True)
    def validate_subclasses(cls, value, field):
        #Create uuid and datarow id so we don't have to define classification objects twice
        #This is caused by the fact that we require these ids for top level classifications but not for subclasses
        results = []
        dummy_id = 'child'.center(25, '_')
        for row in value:
            results.append({
                **row, 'dataRow': {
                    'id': dummy_id
                },
                'uuid': str(uuid4())
            })
        return results


class NDPolygon(NDBaseTool):
    ontology_type: Literal["polygon"] = "polygon"
    polygon: List[Point] = pydantic_compat.Field(determinant=True)

    @pydantic_compat.validator('polygon')
    def is_geom_valid(cls, v):
        if len(v) < 3:
            raise ValueError(
                f"A polygon must have at least 3 points to be valid. Found {v}")
        return v


class NDPolyline(NDBaseTool):
    ontology_type: Literal["line"] = "line"
    line: List[Point] = pydantic_compat.Field(determinant=True)

    @pydantic_compat.validator('line')
    def is_geom_valid(cls, v):
        if len(v) < 2:
            raise ValueError(
                f"A line must have at least 2 points to be valid. Found {v}")
        return v


class NDRectangle(NDBaseTool):
    ontology_type: Literal["rectangle"] = "rectangle"
    bbox: Bbox = pydantic_compat.Field(determinant=True)
    #Could check if points are positive


class NDPoint(NDBaseTool):
    ontology_type: Literal["point"] = "point"
    point: Point = pydantic_compat.Field(determinant=True)
    #Could check if points are positive


class EntityLocation(pydantic_compat.BaseModel):
    start: int
    end: int


class NDTextEntity(NDBaseTool):
    ontology_type: Literal["named-entity"] = "named-entity"
    location: EntityLocation = pydantic_compat.Field(determinant=True)

    @pydantic_compat.validator('location')
    def is_valid_location(cls, v):
        if isinstance(v, pydantic_compat.BaseModel):
            v = v.dict()

        if len(v) < 2:
            raise ValueError(
                f"A line must have at least 2 points to be valid. Found {v}")
        if v['start'] < 0:
            raise ValueError(f"Text location must be positive. Found {v}")
        if v['start'] > v['end']:
            raise ValueError(
                f"Text start location must be less or equal than end. Found {v}"
            )
        return v


class RLEMaskFeatures(pydantic_compat.BaseModel):
    counts: List[int]
    size: List[int]

    @pydantic_compat.validator('counts')
    def validate_counts(cls, counts):
        if not all([count >= 0 for count in counts]):
            raise ValueError(
                "Found negative value for counts. They should all be zero or positive"
            )
        return counts

    @pydantic_compat.validator('size')
    def validate_size(cls, size):
        if len(size) != 2:
            raise ValueError(
                f"Mask `size` should have two ints representing height and with. Found : {size}"
            )
        if not all([count > 0 for count in size]):
            raise ValueError(
                f"Mask `size` should be a postitive int. Found : {size}")
        return size


class PNGMaskFeatures(pydantic_compat.BaseModel):
    # base64 encoded png bytes
    png: str


class URIMaskFeatures(pydantic_compat.BaseModel):
    instanceURI: str
    colorRGB: Union[List[int], Tuple[int, int, int]]

    @pydantic_compat.validator('colorRGB')
    def validate_color(cls, colorRGB):
        #Does the dtype matter? Can it be a float?
        if not isinstance(colorRGB, (tuple, list)):
            raise ValueError(
                f"Received color that is not a list or tuple. Found : {colorRGB}"
            )
        elif len(colorRGB) != 3:
            raise ValueError(
                f"Must provide RGB values for segmentation colors. Found : {colorRGB}"
            )
        elif not all([0 <= color <= 255 for color in colorRGB]):
            raise ValueError(
                f"All rgb colors must be between 0 and 255. Found : {colorRGB}")
        return colorRGB


class NDMask(NDBaseTool):
    ontology_type: Literal["superpixel"] = "superpixel"
    mask: Union[URIMaskFeatures, PNGMaskFeatures,
                RLEMaskFeatures] = pydantic_compat.Field(determinant=True)


#A union with custom construction logic to improve error messages
class NDTool(
        SpecialUnion,
        Type[Union[  # type: ignore
            NDMask,
            NDTextEntity,
            NDPoint,
            NDRectangle,
            NDPolyline,
            NDPolygon,
        ]]):
    ...


class NDAnnotation(
        SpecialUnion,
        Type[Union[  # type: ignore
            NDTool, NDClassification]]):

    @classmethod
    def build(cls: Any, data) -> "NDBase":
        if not isinstance(data, dict):
            raise ValueError('value must be dict')
        errors = []
        for cl in cls.get_union_types():
            try:
                return cl(**data)
            except KeyError as e:
                errors.append(f"{cl.__name__}: {e}")

        raise ValueError('Unable to construct any annotation.\n{}'.format(
            "\n".join(errors)))

    @classmethod
    def schema(cls):
        data = {'definitions': {}}
        for type_ in cls.get_union_types():
            schema_ = type_.schema()
            data['definitions'].update(schema_.pop('definitions'))
            data[type_.__name__] = schema_
        return data
